#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import logging

import numpy as np

# Local imports
from .resample import resample
from .ts_vec_xyz import ts_vec_xyz

# from .e_vxb import e_vxb


__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def vht(e, b, no_ez: bool = False):
    r"""Estimate velocity of the De Hoffman-Teller frame from the velocity
    estimate the electric field eht=-vht x b

    Parameters
    ----------
    e : xarray.DataArray
        Time series of the electric field.
    b : xarray.DataArray
        Time series of the magnetic field.
    no_ez : boolean, Optional
        If True assumed no Ez. Default is False.

    Returns
    -------
    vht : numpy.ndarray
        De Hoffman Teller frame velocity [km/s].
    vht : xarray.DataArray
        Time series of the electric field in the De Hoffman frame.
    dv_ht : numpy.ndarray
        Error of De Hoffman Teller frame.

    """

    n_samples = len(e)

    # Resample magnetic field to electric field sampling (usually higher)

    if n_samples != len(b):
        b = resample(b, e)

    p = np.zeros(6)

    p[0] = np.sum(b[:, 0].data * b[:, 0].data) / n_samples  # Bx*Bx
    p[1] = np.sum(b[:, 0].data * b[:, 1].data) / n_samples  # Bx*By
    p[2] = np.sum(b[:, 0].data * b[:, 2].data) / n_samples  # Bx*Bz
    p[3] = np.sum(b[:, 1].data * b[:, 1].data) / n_samples  # By*By
    p[4] = np.sum(b[:, 1].data * b[:, 2].data) / n_samples  # By*Bz
    p[5] = np.sum(b[:, 2].data * b[:, 2].data) / n_samples  # Bz*Bz

    # assume only Ex and Ey
    if no_ez:
        e[:, 2] *= 0  # put z component to 0 when using only Ex and Ey

        k_mat = np.array(
            [[p[5], 0, -p[2]], [0, p[5], -p[4]], [-p[2], -p[4], p[0] + p[3]]],
        )
    else:
        k_mat = np.array(
            [
                [p[3] + p[5], -p[1], -p[2]],
                [-p[1], p[0] + p[5], -p[4]],
                [-p[2], -p[4], p[0] + p[3]],
            ],
        )

    exb = np.cross(e, b)

    ind_data = np.where(~np.isnan(exb[:, 0].data))[0]

    # revised by Wenya LI; 2015-11-21, wyli @ irfu
    exb_avg = np.sum(exb[ind_data, :], axis=0) / n_samples

    # averExB=sum(ExB(indData).data,1)/nSamples;
    # end revise.
    v_ht = np.linalg.solve(k_mat, exb_avg.T) * 1e3  # 9.12 in ISSI book

    v_ht_hat = v_ht / np.linalg.norm(v_ht, keepdims=True)

    logging.info(
        "v_ht =%(v_mag)7.4f * %(v_vec)s km/s",
        {"v_mag": np.linalg.norm(v_ht), "v_vec": np.array_str(v_ht_hat)},
    )

    # Calculate the goodness of the Hoffman Teller frame
    # e_ht = e_vxb(v_ht, b)
    e_ht = ts_vec_xyz(b.time.data, -1e-3 * np.cross(v_ht, b.data))

    if no_ez:
        e_p, e_ht_p = [e[ind_data], e_ht[ind_data]]
        e_p.data[:, 2], e_ht_p.data[:, 2] = [0, 0]
    else:
        e_p, e_ht_p = [e[ind_data], e_ht[ind_data]]

    delta_e = e_p.data - e_ht_p.data

    poly_fit = np.polyfit(
        e_ht_p.data.reshape([len(e_ht_p) * 3]),
        e_p.data.reshape([len(e_p) * 3]),
        1,
    )
    corr_coeff = np.corrcoef(
        e_ht_p.data.reshape([len(e_ht_p) * 3]),
        e_p.data.reshape([len(e_p) * 3]),
    )

    logging.info(
        "slope = %(slope)6.4f, offs = %(offset)6.4f, cc = %(cc)6.4f",
        {"slope": poly_fit[0], "offset": poly_fit[1], "cc": corr_coeff[0, 1]},
    )

    dv_ht = np.sum(np.sum(delta_e**2)) / len(ind_data)
    s_mat = (dv_ht / (2 * len(ind_data) - 3)) / k_mat
    dv_ht = np.sqrt(np.diag(s_mat)) * 1e3

    dv_ht_hat = dv_ht / np.linalg.norm(dv_ht)

    logging.info(
        "dv_ht =%(dv_mag)7.4f * %(dv_vec)s km/s",
        {"dv_mag": np.linalg.norm(dv_ht), "dv_vec": np.array_str(dv_ht_hat)},
    )

    return v_ht, e_ht, dv_ht



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
from scipy import constants
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["dpf2psd"]


def _mass_ratio(inp: Union[Dataset, DataArray]) -> float:
    r"""Compute mass ratio of the input species.

    Parameters
    ----------
    inp : Dataset or DataArray
        Input distribution function.

    Returns
    -------
    float
        Mass ratio of the species.

    Raises
    ------
    ValueError
        If the species is not supported.

    """
    if inp.attrs["species"].lower() in ["ions", "ion", "protons", "proton"]:
        mass_ratio = 1
    elif inp.attrs["species"].lower() in ["alphas", "alpha", "helium"]:
        mass_ratio = 4
    elif inp.attrs["species"].lower() in ["electrons", "e"]:
        mass_ratio = constants.electron_mass / constants.proton_mass
    else:
        raise ValueError("Invalid specie")

    return mass_ratio


def _convert(inp: np.ndarray, units: str, mass_ratio: float) -> np.ndarray:
    r"""Convert differential particle flux to phase space density.

    Parameters
    ----------
    inp : np.ndarray
        Input differential particle flux.
    units : str
        Units of the input differential particle flux.
    mass_ratio : float
        Mass ratio of the species.

    Returns
    -------
    tmp_data : np.ndarray
        Phase space density data.

    Raises
    ------
    ValueError
        If the input unit is not supported.

    """
    fact = 1e6 * 0.53707 * mass_ratio**2

    if units.lower() == "1/(cm^2 s sr kev)":
        tmp_data = inp * 1e-3 / 1e18 * fact
    elif units.lower() == "1/(cm^2 s sr ev)":
        tmp_data = inp / 1e18 * fact
    else:
        raise ValueError("Invalid unit")

    return tmp_data


def dpf2psd(inp: Union[Dataset, DataArray]) -> Union[Dataset, DataArray]:
    r"""Compute phase space density from differential particle flux.

    Parameters
    ----------
    inp : DataArray or Dataset
        Time series of the differential particle flux in
        [(cm^{2} s sr keV)^{-1}].

    Returns
    -------
    psd : DataArray or Dataset
        Time series of the phase space density in [s^{3} m^{-6}].

    Raises
    ------
    TypeError
        If the input type is not supported.

    """
    if isinstance(inp, Dataset):
        tmp_data = _convert(inp.data.data, inp.data.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy[:, :, None, None], (1, 1, *tmp_data.shape[2:]))
        tmp_data /= energy_mat
        out = inp.copy()
        out.data.data = np.squeeze(tmp_data)
        out.data.attrs["UNITS"] = "s^3/m^6"
    elif isinstance(inp, DataArray):
        tmp_data = _convert(inp.data, inp.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy, (tmp_data.shape[0], 1))
        tmp_data /= energy_mat
        out = inp.copy()
        out.data = np.squeeze(tmp_data)
        out.attrs["UNITS"] = "s^3/m^6"
    else:
        raise TypeError("Invalid input type")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def t_eval(inp, times):
    r"""Evaluates the input time series at the target time.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series if the input to evaluate.
    times : ndarray
        Times at which the input will be evaluated.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input at times t.

    """

    idx = np.zeros(len(times))

    for i, time in enumerate(times):
        idx[i] = bisect.bisect_left(inp.time.data, time)

    idx = idx.astype(np.int64)

    if inp.ndim == 2:
        out = xr.DataArray(
            inp.data[idx, :],
            coords=[times, inp.comp],
            dims=["time", "comp"],
        )
    else:
        out = xr.DataArray(inp.data[idx], coords=[times], dims=["time"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .ts_scalar import ts_scalar

__author__ = "Apostolos Kolokotronis"
__email__ = "apostolos.kolokotronis@irf.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def sliding_derivative(
    time_series, t_units: str = "ns", window_size=3, method: str = "window"
):
    """
    Compute the sliding time derivative of a time series using central differences.

    This function calculates the rate of change of a time series at each point by
    computing the derivative within a sliding window of a specified size. The derivative
    at each point is estimated using the first and last points in the window (central
    difference).

    Parameters:
    ----------
    time_series : xarray.DataArray
        The time series data for which the derivative is to be calculated.
    t_units: str, optional, default: "ns"
        The units of the time coordinate in time_series. If t_units is "ns" then the
        time is converted to seconds.
    window_size : int, optional, default: 3
        The number of data points in each sliding window. It should be an odd integer
        to ensure a symmetric window around the central point for central differences.

    Returns:
    -------
    derivative : xarray.DataArray
        An array containing the sliding derivative for each point in the time series.
        The length of this array matches the length of `time_series`. For points near
        the boundaries (where a full window cannot be formed), the result will be NaN.

    Notes:
    ------
    - The derivative is approximated using central differences for points that can
    accommodate the window size. For edge points, the output will contain NaN.
    - The function assumes `time_steps` are evenly spaced but works for irregular time
    steps as well by calculating the actual time difference between the start and end
    of the window.

    """

    data = time_series.data
    time = time_series.time.astype(np.float64).data

    assert t_units.lower() in ["ns", "s"], "convert time to ns or s."

    if t_units.lower() == "ns" or time.data.dtype == "<M8[ns]":
        time = time * 1e-9

    assert method.lower() in [
        "window",
        "5ps",
        "9ps",
    ], "this method has not been implemented."

    half_window = window_size // 2

    # Fill the output with NaN for edge cases
    derivative = np.full(len(data), np.nan)
    if method == "window":

        half_window = window_size // 2

        # Iterate over each window in the time series
        for i in range(half_window, len(data) - half_window):
            # Get the window of values and time
            values_window = data[i - half_window : i + half_window + 1]
            time_window = time[i - half_window : i + half_window + 1]

            # Compute finite differences (central difference for the middle point)
            derivative[i] = (values_window[-1] - values_window[0]) / (
                time_window[-1] - time_window[0]
            )
    elif method == "5ps":

        for i in range(2, len(data) - 2):

            derivative[i] = (
                1 / 12 * data[i - 2]
                - 2 / 3 * data[i - 1]
                + 0 * data[i]
                + 2 / 3 * data[i + 1]
                - 1 / 12 * data[i + 2]
            ) / ((time[i + 2] - time[i - 2]) * 0.25)
    elif method == "9ps":

        for i in range(4, len(data) - 4):

            derivative[i] = (
                1 / 280 * data[i - 4]
                - 4 / 105 * data[i - 3]
                + 1 / 5 * data[i - 2]
                - 4 / 5 * data[i - 1]
                + 0 * data[i]
                + 4 / 5 * data[i + 1]
                - 1 / 5 * data[i + 2]
                + 4 / 105 * data[i + 3]
                - 1 / 280 * data[i + 4]
            ) / ((time[i + 4] - time[i - 4]) * 1 / 8)

    # time_dt64 = ts_time(time).data
    out = ts_scalar(
        time_series.time.data,
        derivative,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

EMPIRICAL_FITS = {
    "10^-2": {
        "proton-cyclotron": {"a": 0.69, "b": 0.4, "beta0": 0.0},
        "mirror": {"a": 1.040, "b": 0.633, "beta0": -0.012},
        "parallel-firehose": {"a": -0.647, "b": 0.583, "beta0": 0.713},
        "oblique-firehose": {"a": -1.447, "b": 1, "beta0": -0.148},
    }
}


def ion_anisotropy_thresh(beta, instability_type, growth_rate="10^-2"):
    r"""Compute the threshold using the empirical model defined in [1]_.

    .. math::

        R_i = T_{i\perp} / T_{i\parallel} = 1 + a / (\beta_{i\parallel} -\beta_0)^b


    Parameters
    ----------
    beta : xarray.DataArray or numpy.ndarray
        Time series or array of parallel ion plasma beta.
    a : float
        Coefficient from fit.
    b : float
        Coefficient from fit.
    beta0 : float
        Coefficient from fit.


    Returns
    -------
    r_i_thresh : xarray.DataArray
        Time series or array of threshold temperature anisotropy at the given beta.


    References
    ----------
    .. [1]  Hellinger, P., P. Travnicek, J. C. Kasper, and A. J. Lazarus (2006),
            Solar wind proton temperature anisotropy: Linear theory and WIND/SWE
            observations, Geophys. Res. Lett., 33, L09101, doi:10.1029/2006GL025925.

    """

    if growth_rate not in EMPIRICAL_FITS:
        raise ValueError(f"Growth rate {growth_rate} not recognized.")

    if instability_type not in EMPIRICAL_FITS[growth_rate]:
        raise ValueError(f"Instability type {instability_type} not recognized.")

    # Get fit parameters
    fit_params = EMPIRICAL_FITS[growth_rate][instability_type]
    a = fit_params["a"]
    b = fit_params["b"]
    beta0 = fit_params["beta0"]

    # Discard negative values from the denominator
    beta[beta < beta0] = np.nan

    # Compute threshold
    r_i_thresh = 1 + a / (beta - beta0) ** b

    return r_i_thresh



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eis_pad_combine_sc(pads):
    r"""Generate composite Pitch Angle Distributions (PAD) from the EIS
    sensors across the MMS spacecraft.

    Parameters
    ----------
    pads : list of xarray.DataArray
        Pitch-angle distribution for all spacecrafts.

    Returns
    -------
    allmms_pad_avg : xarray.DataArray
        Composite pitch angle distribution.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_pad, pyrfu.mms.eis_spec_combine_sc

    """
    # Determine spacecraft with smallest number of time steps to use as
    # reference spacecraft
    time_size = [len(probe.time.data) for probe in pads]
    ref_sc_loc = np.argmin(time_size)
    ref_probe = pads[ref_sc_loc]

    # Define common energy grid across EIS instruments
    n_en_chans = [len(probe.energy.data) for probe in pads]

    size_en = np.min(n_en_chans)
    energy_data = [probe.energy.data[:size_en] for probe in pads]
    energy_data = np.stack(energy_data)
    common_energy = np.nanmean(energy_data, axis=0)

    # create PA labels
    n_pabins = len(ref_probe.theta.data)
    size_pabin = 180 / n_pabins
    pa_label = 180.0 * np.arange(n_pabins) / n_pabins + size_pabin / 2.0

    allmms_pad = np.zeros(
        (
            ref_probe.shape[0],
            ref_probe.shape[1],
            ref_probe.shape[2],
            len(pads),
        ),
    )

    for i_pad, pad_ in enumerate(pads):
        allmms_pad[..., i_pad] = pad_.data[: len(ref_probe.time), ...]

    allmms_pad_avg = np.nanmean(allmms_pad, axis=3)

    allmms_pad_avg = xr.DataArray(
        allmms_pad_avg,
        coords=[ref_probe.time.data, pa_label, common_energy],
        dims=["time", "theta", "energy"],
    )

    return allmms_pad_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Any, Union

# 3rd party imports
import numpy as np
from numpy.typing import NDArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def unix2datetime64(time: Union[list[float], NDArray[Any]]) -> NDArray[np.datetime64]:
    r"""Converts unix time to datetime64 in ns units.

    Parameters
    ----------
    time : numpy.ndarray
        Time in unix format.

    Returns
    -------
    time_datetime64 : numpy.ndarray
        Time in datetime64 format.

    Raises
    ------
    TypeError
        If time is not a list or numpy.ndarray.

    See Also
    --------
    pyrfu.pyrf.datetime642unix

    """
    # Check input type
    if isinstance(time, (list, np.ndarray)):
        time_array = np.array(time)
    else:
        raise TypeError("time must be list or numpy.ndarray")

    # Make sure that time is in ns format
    time_unix = (time_array * 1e9).astype(np.int64)

    # Convert to unix
    time_datetime64 = time_unix.astype("datetime64[ns]")

    return time_datetime64



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# #rd party imports
import numpy as np
import xarray as xr
from scipy import interpolate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_pad_spinavg(pad, spin_sectors, bin_size: float = 16.3636):
    r"""Spin-average the FEEPS pitch angle distributions.

    Parameters
    ----------
    pad : xarray.DataArray
        Pitch angle distribution.
    spin_sectors : xarray.DataArray
        Time series of the spin sectors.
    bin_size : float, Optional
        Size of the pitch angle bins

    Returns
    -------
    out : xarray.DataArray
        Spin averaged pitch angle distribution.

    """

    n_pabins = int(180.0 / bin_size)
    new_bins = 180.0 * np.arange(int(n_pabins + 1)) / n_pabins

    # get the spin sectors
    # v5.5+ = mms1_epd_feeps_srvy_l1b_electron_spinsectnum
    spin_sectors = spin_sectors.data

    spin_starts = np.where(spin_sectors[:-1] >= spin_sectors[1:])[0] + 1

    times = pad.time.data
    data = pad.data
    angles = pad.theta.data

    n_spin = len(spin_starts)
    n_angs = len(angles)

    spin_avg_flux = np.zeros([n_spin, n_angs])
    rebinned_data = np.zeros([n_spin, int(n_pabins + 1)])
    spin_times = np.zeros(n_spin, dtype="<M8[ns]")

    # the following is for rebinning and interpolating to new_bins
    srx = n_angs / (n_pabins + 1) * (np.arange(int(n_pabins + 1)) + 0.5) - 0.5

    c_start = 0
    for i, spin in enumerate(spin_starts):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            spin_avg_flux[i, :] = np.nanmean(
                data[c_start : spin + 1, :],
                axis=0,
            )
            spin_times[i] = times[c_start]

            # rebin and interpolate to new_bins
            spin_avg_interp = interpolate.interp1d(
                np.arange(n_angs),
                spin_avg_flux[i, :],
                fill_value="extrapolate",
            )
            rebinned_data[i, :] = spin_avg_interp(srx)

            # we want to take the end values instead of extrapolating
            rebinned_data[i, 0] = spin_avg_flux[i, 0]
            rebinned_data[i, -1] = spin_avg_flux[i, -1]

        c_start = spin + 1

    out = xr.DataArray(
        rebinned_data,
        coords=[spin_times, new_bins],
        dims=["time", "theta"],
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import signal

from .calc_dt import calc_dt
from .filt import filt
from .integrate import integrate
from .norm import norm

# Local imports
from .resample import resample
from .ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def match_phibe_dir(b_xyz, e_xyz, angles: np.ndarray = None, f: float = None):
    r"""Get propagation direction by matching dBpar and "phi". Tries different
    propagation directions and finds the direction perpendicular to the
    magnetic field that gives the best correlation between the electrostatic
    potential and the parallel wave magnetic field according to

    .. math::

            \int E \textrm{d}t = \frac{B_0}{ne \mu_0} B_{wave}


    Parameters
    ----------
    b_xyz : xarray.DataArray
        Time series of the magnetic field (to be filtered if f is given).
    e_xyz : xarray.DataArray
        Time series of the electric field (to be filtered if f is given).
    angles : array_like, Optional
        The angles in degrees to try (1-180 default)
    f : float, Optional
        Filter frequency.

    Returns
    -------
    x : ndarray
        Normal direction (size: n_triesx3).
    y : ndarray
        Propagation direction.
    z : ndarray
        Magnetic field direction.
    corr_vec : ndarray
        Correlation vector.
    int_e_dt : ndarray
        Potential.
    b_z : ndarray
        Wave magnetic field in parallel direction.
    b_0 : ndarray
        Mean magnetic field.
    de_k : ndarray
        Wave electric field in propagation direction.
    de_n : ndarray
        Wave electric field in propagation normal direction.
    e_k : ndarray
        Electric field in propagation direction.
    e_n : ndarray
        Electric field in propagation normal direction.

    """

    # Resample B to E if they have different size
    b_xyz = resample(b_xyz, e_xyz)

    # Filter if f is given, otherwise assume it is filtered
    if f is not None:
        b_ac = filt(b_xyz, f, 0, 5)
        e_ac = filt(e_xyz, f, 0, 5)
    else:
        b_ac = b_xyz
        e_ac = e_xyz

    # Get background magnetic field, for irf_match_phibe_v
    b_0 = np.mean(norm(b_xyz))

    # If no angles are specified, set 1,4,7,...,158 as default
    if angles is None:
        angles = np.linspace(0, 360, 121)

    # number of angles
    na_ = len(angles)

    # Set up coordinate systems
    b_hat = np.mean(b_xyz.data, axis=0)
    b_hat /= np.linalg.norm(b_hat)
    y_ = np.cross(np.cross(b_hat, np.array([1, 0, 0])), b_hat)
    y_ /= np.linalg.norm(y_)
    x_ = np.cross(y_, b_hat)
    x_ /= np.linalg.norm(x_)
    x_ = np.tile(x_, (na_, 1))  # perp1
    y_ = np.tile(y_, (na_, 1))  # perp2
    z_ = np.tile(b_hat, (na_, 1))  # B / z direction, tries * 3

    theta = np.linspace(0, 2 * np.pi - np.pi / na_, na_)  # angles

    x_n = x_ * np.transpose(np.tile(np.cos(theta), (3, 1)))
    x_n += y_ * np.transpose(np.tile(np.sin(theta), (3, 1)))

    y_ = np.cross(z_, x_n)
    x_ = x_n

    # Field aligned B
    b_z = np.sum(b_ac * z_[0, :], axis=1)

    # Allocate correlations
    corr_vec = np.zeros(na_)

    # Allocate vectors, 4 first used for illustration
    int_e_dt = np.zeros((len(e_xyz), na_))  # potential
    e_k, e_n = [np.zeros((len(e_xyz), na_)) for _ in range(2)]  #
    de_k, de_n = [np.zeros((len(e_xyz), na_)) for _ in range(2)]  #

    # Integrate E in all x - directions
    for k in range(na_):
        de_k[:, k] = np.sum(e_ac.data * x_[k, :], axis=1)
        de_n[:, k] = np.sum(e_ac.data * y_[k, :], axis=1)

        e_k[:, k] = np.sum(e_xyz.data * x_[k, :], axis=1)
        e_n[:, k] = np.sum(e_xyz.data * y_[k, :], axis=1)

        # Get Phi_E = int(Ek), there's no minus since the field is integrated
        # in the opposite direction of the wave propagation direction.
        prel_ = integrate(
            ts_scalar(e_xyz.time.data, de_k[:, k]),
            calc_dt(e_xyz),
        )
        int_e_dt[:, k] = prel_.data - np.mean(prel_.data)

        # Get correlation
        corr_ = signal.correlate(int_e_dt[:, k], b_z[:, 1], mode="valid")
        scale = np.sqrt(
            np.dot(int_e_dt[:, k], int_e_dt[:, k]) * np.dot(b_z[:, 1], b_z[:, 1]),
        )
        corr_vec[k] = corr_ / scale

    return x_, y_, z_, corr_vec, int_e_dt, b_z, b_0, de_k, de_n, e_k, e_n



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import random
import unittest

import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np
from ddt import data, ddt, unpack
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.image import AxesImage

# Local imports
from .. import plot
from . import generate_data, generate_ts


@ddt
class PlotLineTestCase(unittest.TestCase):
    @data((0.0, generate_ts(64.0, 100)), (plt.subplots(3)[1], generate_ts(64.0, 100)))
    @unpack
    def test_plot_line_axis_type(self, axis, inp):
        with self.assertRaises(TypeError):
            plot.plot_line(axis, inp)

    @data((plt.subplots(1)[1], generate_data(100)))
    @unpack
    def test_plot_line_inp_type(self, axis, inp):
        with self.assertRaises(TypeError):
            plot.plot_line(axis, inp)

    @data(
        (plt.subplots(1)[1], generate_ts(64.0, 100, tensor_order=random.randint(3, 10)))
    )
    @unpack
    def test_plot_line_inp_shape(self, axis, inp):
        with self.assertRaises(NotImplementedError):
            plot.plot_line(axis, inp)

    @data(
        (None, generate_ts(64.0, 100, tensor_order=random.randint(0, 2))),
        (plt.subplots(1)[1], generate_ts(64.0, 100, tensor_order=random.randint(0, 2))),
        (
            plt.subplots(3)[1][0],
            generate_ts(64.0, 100, tensor_order=random.randint(0, 2)),
        ),
    )
    @unpack
    def test_plot_line_output(self, axis, inp):
        result = plot.plot_line(axis, inp)
        self.assertIsInstance(result, Axes)


@ddt
class AddPositionTestCase(unittest.TestCase):
    @data(generate_ts(64.0, 100, tensor_order=1))
    def test_add_position_output(self, value):
        result = plot.add_position(plt.subplots(1)[1], value)
        self.assertIsInstance(result, Axes)


@ddt
class PlTxTestCase(unittest.TestCase):
    @data(
        ([generate_ts(64.0, 100, tensor_order=3) for _ in range(4)], "cluster"),
        ([generate_ts(64.0, 100, tensor_order=0) for _ in range(4)], "bazinga"),
    )
    @unpack
    def test_pl_tx_input(self, value, colors):
        with self.assertRaises(NotImplementedError):
            plot.pl_tx(plt.subplots(1)[1], value, colors=colors)

    @data(
        (
            None,
            [
                generate_ts(64.0, 100, tensor_order=0),
                generate_ts(64.0, 100, tensor_order=0),
                generate_ts(64.0, 100, tensor_order=0),
                generate_ts(64.0, 100, tensor_order=0),
            ],
        ),
        (
            plt.subplots(1)[1],
            [
                generate_ts(64.0, 100, tensor_order=0),
                generate_ts(64.0, 100, tensor_order=0),
                generate_ts(64.0, 100, tensor_order=0),
                generate_ts(64.0, 100, tensor_order=0),
            ],
        ),
        (
            plt.subplots(1)[1],
            [
                generate_ts(64.0, 100, tensor_order=1),
                generate_ts(64.0, 100, tensor_order=1),
                generate_ts(64.0, 100, tensor_order=1),
                generate_ts(64.0, 100, tensor_order=1),
            ],
        ),
        (
            plt.subplots(1)[1],
            [
                generate_ts(64.0, 100, tensor_order=2),
                generate_ts(64.0, 100, tensor_order=2),
                generate_ts(64.0, 100, tensor_order=2),
                generate_ts(64.0, 100, tensor_order=2),
            ],
        ),
    )
    @unpack
    def test_pl_tx_output(self, ax, value):
        result = plot.pl_tx(ax, value, 0)
        self.assertIsInstance(result, Axes)


@ddt
class ZoomTestCase(unittest.TestCase):
    @data((None, plt.subplots(2)[1][1]), (plt.subplots(2)[1][0], None))
    @unpack
    def test_zoom_input(self, ax1, ax2):
        with self.assertRaises(TypeError):
            plot.zoom(ax1, ax2)

    @data((plt.subplots(2)[1][0], plt.subplots(2)[1][1]))
    @unpack
    def test_zoom_output(self, ax1, ax2):
        result = plot.zoom(ax1, ax2)
        self.assertIsInstance(result, tuple)
        self.assertIsInstance(result[0], Axes)
        self.assertIsInstance(result[1], Axes)


@ddt
class SetColorCycleTestCase(unittest.TestCase):
    @data("pyrfu", "oceanic", "tab", "")
    def set_color_cycle_input(self, value):
        result = plot.set_color_cycle(value)
        self.asssertIsInstance(result[0], list)
        self.asssertIsInstance(result[1], str)


@ddt
class PlotHeatmapTestCase(unittest.TestCase):
    @data(
        (plt.subplots(1)[1], "bazinga", np.random.rand(10), np.random.rand(10)),
        (plt.subplots(1)[1], np.random.rand(10, 10), "bazinga", np.random.rand(10)),
        (plt.subplots(1)[1], np.random.rand(10, 10), np.random.rand(10), "bazinga"),
    )
    @unpack
    def test_plot_heatmap_input_types(self, ax, data, x, y):
        with self.assertRaises(TypeError):
            plot.plot_heatmap(ax, data, x, y)

    @data(
        (
            plt.subplots(1)[1],
            np.random.rand(10, 10),
            np.random.rand(9),
            np.random.rand(10),
        ),
        (
            plt.subplots(1)[1],
            np.random.rand(10, 10),
            np.random.rand(10),
            np.random.rand(9),
        ),
    )
    @unpack
    def test_plot_heatmap_input_shape(self, ax, data, x, y):
        with self.assertRaises(ValueError):
            plot.plot_heatmap(ax, data, x, y)

    @data(
        (None, np.random.rand(10, 10), np.random.rand(10), np.random.rand(10)),
        (
            plt.subplots(1)[1],
            np.random.rand(10, 10),
            np.random.rand(10),
            np.random.rand(10),
        ),
    )
    @unpack
    def test_plot_heatmap_output(self, ax, data, x, y):
        result = plot.plot_heatmap(ax, data, x, y)
        self.assertIsInstance(result[0], AxesImage)
        self.assertIsInstance(result[1], Colorbar)


@ddt
class AnnotateHeatmapTestCase(unittest.TestCase):
    def test_annotate_heatmap_input(self):
        _, ax = plt.subplots(1)
        # Create image
        im, _ = plot.plot_heatmap(
            ax, np.random.rand(10, 10), np.random.rand(10), np.random.rand(10)
        )
        with self.assertRaises(TypeError):
            plot.annotate_heatmap(im, "bazinga")

    @data(
        (
            plt.subplots(1)[1],
            np.random.rand(10, 10),
            np.random.rand(10),
            np.random.rand(10),
            random.random(),
        ),
        (
            plt.subplots(1)[1],
            np.random.rand(10, 10),
            np.random.rand(10),
            np.random.rand(10),
            None,
        ),
    )
    @unpack
    def test_annotate_heatmap_output(self, ax, data, x, y, threshold):
        # Create image
        im, _ = plot.plot_heatmap(ax, data, x, y)

        # Test with data provided
        plot.annotate_heatmap(im, data, threshold=threshold)

        # Test with no data provided
        plot.annotate_heatmap(im)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import datetime
import fnmatch
import glob
import json
import os
import re
from typing import Optional, Union

# 3rd party imports
import numpy as np
import pandas as pd

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.iso86012datetime import iso86012datetime
from ..pyrf.iso86012datetime64 import iso86012datetime64
from .db_init import MMS_CFG_PATH

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def list_files_ancillary(
    tint: list[str],
    mms_id: Union[str, int],
    product: str,
    data_path: Optional[str] = "",
) -> list[str]:
    r"""Find available ancillary files in the data directories for the target product
    type.

    Parameters
    ----------
    tint : list of str
        Time interval
    mms_id : str or int
        Spacecraft index
    product : str
        Ancillary type: {"predatt", "predeph", "defatt", "defeph"}.
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu.mms.mms_config.py`

    Returns
    -------
    file_names : list
        Ancillary files in interval.

    """

    # Check path
    if not data_path:
        # Read the current version of the MMS configuration file
        with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
            config = json.load(fs)

        root_path = os.path.normpath(config["local"])
    else:
        root_path = os.path.normpath(data_path)

    # Make sure that the data path exists
    assert os.path.exists(root_path), f"{root_path} doesn't exist!!"

    if isinstance(mms_id, int):
        mms_id = str(mms_id)

    # Check time interval
    if isinstance(tint, list):
        tint_array = np.array(tint)
    else:
        raise TypeError("tint must be a list!!")

    # Convert time interval to ISO 8601
    if isinstance(tint_array[0], str):
        tint_iso8601 = datetime642iso8601(iso86012datetime64(tint_array))
        tint_datetime = iso86012datetime(tint_iso8601)
    else:
        raise TypeError("Values must be in str!!")

    # PAD time interval to handle ancillary file start after midnight
    tint_datetime[0] = tint_datetime[0] - datetime.timedelta(days=1)
    # tint_datetime[1] = tint_datetime[1] + datetime.timedelta(days=1)

    # directory and file name search patterns
    # For now
    # -all ancillary data is in one directory:
    #       mms\ancillary
    # -assume file names are of the form:
    #   SPACECRAFT_FILETYPE_startDate_endDate.version
    #   where SPACECRAFT is [MMS1, MMS2, MMS3, MMS4] in uppercase
    #   and FILETYPE is either DEFATT, PREDATT, DEFEPH, PREDEPH in uppercase
    #   and start/endDate is YYYYDOY
    #   and version is Vnn (.V00, .V01, etc..)
    dir_pattern = os.sep.join([root_path, "ancillary", f"mms{mms_id}", product])
    file_pattern = "_".join([f"MMS{mms_id}", product.upper(), "???????_???????.V??"])

    # find all files in the directory
    files = glob.glob(os.sep.join([dir_pattern, file_pattern]))

    # find the files within the time interval
    fname_fmt = f"MMS{mms_id}_{product.upper()}_([0-9]{{7}})_([0-9]{{7}}).V[0-9]{{2}}"

    if os.name == "nt":
        full_path = os.sep.join([re.escape(dir_pattern) + os.sep, fname_fmt])
    else:
        full_path = os.sep.join([re.escape(dir_pattern), fname_fmt])

    file_regex = re.compile(full_path)

    files_in_tint = []

    for file in files:
        time_match = file_regex.match(file)
        if time_match is not None:
            start_time = pd.to_datetime(time_match.group(1), format="%Y%j")
            end_time = pd.to_datetime(time_match.group(2), format="%Y%j")
            if start_time < tint_datetime[1] and end_time >= tint_datetime[0]:
                files_in_tint.append(file)

    # ensure only the latest version of each file is loaded
    out_files = []

    for file_in_tint in files_in_tint:
        this_file = file_in_tint[0:-3] + "V??"
        versions = fnmatch.filter(files_in_tint, this_file)
        if len(versions) > 1:
            # only grab the latest version
            out_files.append(sorted(versions)[-1])
        else:
            out_files.append(versions[0])

    file_names = list(set(out_files))
    file_names.sort()

    return file_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

from ..pyrf.resample import resample

# Local imports
from .feeps_active_eyes import feeps_active_eyes

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

# Rotation matrices for FEEPS coord system (FCS) into body coordinate system
# (BCS):
t_top = np.array(
    [
        [1.0 / np.sqrt(2.0), -1.0 / np.sqrt(2.0), 0],
        [1.0 / np.sqrt(2.0), 1.0 / np.sqrt(2.0), 0],
        [0, 0, 1],
    ],
)
t_bot = np.array(
    [
        [-1.0 / np.sqrt(2.0), -1.0 / np.sqrt(2.0), 0],
        [-1.0 / np.sqrt(2.0), 1.0 / np.sqrt(2.0), 0],
        [0, 0, -1],
    ],
)

# the following 2 hash tables map TOP/BOTTOM telescope #s to index of the
# PA array created above
top_tele_idx_map = {
    1: 0,
    2: 1,
    3: 2,
    4: 3,
    5: 4,
    6: 0,
    7: 1,
    8: 2,
    9: 5,
    10: 6,
    11: 7,
    12: 8,
}

bot_tele_idx_map = {
    1: 9,
    2: 10,
    3: 11,
    4: 12,
    5: 13,
    6: 3,
    7: 4,
    8: 5,
    9: 14,
    10: 15,
    11: 16,
    12: 17,
}

# Telescope vectors in FCS:
v_fcs = {
    1: [0.347, -0.837, 0.423],
    2: [0.347, -0.837, -0.423],
    3: [0.837, -0.347, 0.423],
    4: [0.837, -0.347, -0.423],
    5: [-0.087, 0.000, 0.996],
    6: [0.104, 0.180, 0.978],
    7: [0.654, -0.377, 0.656],
    8: [0.654, -0.377, -0.656],
    9: [0.837, 0.347, 0.423],
    10: [0.837, 0.347, -0.423],
    11: [0.347, 0.837, 0.423],
    12: [0.347, 0.837, -0.423],
}

sensor_ids = {
    "electron": {
        "top": [1, 2, 3, 4, 5, 9, 10, 11, 12],
        "bot": [1, 2, 3, 4, 5, 9, 10, 11, 12],
    },
    "ion": {"top": [6, 7, 8], "bot": [6, 7, 8]},
}


def _calc_pas(d_type, b_bcs):
    n_t = len(b_bcs.time)
    n_scopes = len(sensor_ids[d_type]["top"])

    vt_bcs = {i: -np.matmul(t_top, v_fcs[i]) for i in range(1, 13)}
    vb_bcs = {i: -np.matmul(t_bot, v_fcs[i]) for i in range(1, 13)}

    pas = np.empty([n_t, int(2 * n_scopes)])
    for i in range(int(2 * n_scopes)):
        if i < n_scopes:
            v_bcs = vt_bcs[sensor_ids[d_type]["top"][i % n_scopes]]
        else:
            v_bcs = vb_bcs[sensor_ids[d_type]["bot"][i % n_scopes]]

        v_dot_b = np.sum(v_bcs * b_bcs, axis=1)
        v_dot_b /= np.linalg.norm(v_bcs) * np.linalg.norm(b_bcs.data, axis=1)
        pas[:, i] = np.rad2deg(np.arccos(v_dot_b))

    return pas


def _calc_new_pas(pas, b_bcs, eyes):
    n_t = len(b_bcs.time.data)
    n_top = len(eyes["top"])
    n_bot = len(eyes["bottom"])

    top_idxs, bot_idxs = [[], []]

    new_pas = np.empty([n_t, n_top + n_bot])

    for j, top_eye in enumerate(eyes["top"]):
        new_pas[:, j] = pas[:, top_tele_idx_map[top_eye]]
        top_idxs.append(j)

    for j, bot_eye in enumerate(eyes["bottom"]):
        new_pas[:, j + n_top] = pas[:, bot_tele_idx_map[bot_eye]]
        bot_idxs.append(j + n_top)

    return new_pas, top_idxs, bot_idxs


def feeps_pitch_angles(inp_dataset, b_bcs):
    r"""Computes the FEEPS pitch angles for each telescope from magnetic field
    data.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset of the time series of the energy spectrum for each eye of
        FEEPS telescopes.
    b_bcs : xarray.DataArray
        Time series of the magnetic in spacecraft coordinates system.

    Returns
    -------
    out : xarray.DataArray
        Time series of the pitch angles.
    idx_maps : dict
        to fill.

    """

    # get the times from the currently loaded FEEPS data
    times = inp_dataset.time.data

    d_type = inp_dataset.attrs["dtype"]
    d_rate = inp_dataset.attrs["tmmode"]
    mms_id = inp_dataset.attrs["mmsId"]

    tint = np.datetime_as_string(
        np.hstack([np.min(times), np.max(times)]),
        "ns",
    )

    eyes = feeps_active_eyes(inp_dataset.attrs, list(tint), mms_id)

    idx_maps = None

    # pitch angles for each eye at each time
    pas = _calc_pas(d_type, b_bcs)

    if d_type == "electron" and d_rate == "brst":
        new_pas = pas
    else:
        # PAs for only active eyes
        # pitch angles for each eye at eaceh time
        new_pas, top_idxs, bot_idxs = _calc_new_pas(pas, b_bcs, eyes)
        idx_maps = {f"{d_type}-top": top_idxs, f"{d_type}-bottom": bot_idxs}

    out = xr.DataArray(
        new_pas,
        coords=[b_bcs.time.data, np.arange(new_pas.shape[1])],
        dims=["time", "idx"],
    )
    out = resample(out, inp_dataset.time)

    return out, idx_maps



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _print_header():
    print("=" * 70)
    print("IRFU plasma calculator, relativistic effects not fully included")
    print("velocities, gyroradia are relativistically correct")
    print(
        "can somebody fix relativstically correct frequencies Fpe, Fce,.. ?",
    )
    print("=" * 70)


def _print_frequencies(f_pe, f_ce, f_uh, f_lh, f_pp, f_cp, f_col):
    print("\nFrequencies: ")
    print("*" * 12)
    print(f"{'F_pe':>5} = {f_pe:>6.2E} Hz")
    print(f"{'F_ce':>5} = {f_ce:>6.2E} Hz")
    print(f"{'F_uh':>5} = {f_uh:>6.2E} Hz")
    print(f"{'F_lh':>5} = {f_lh:>6.2E} Hz")
    print(f"{'F_pp':>5} = {f_pp:>6.2E} Hz")
    print(f"{'F_cp':>5} = {f_cp:>6.2E} Hz")
    print(f"{'F_col':>5} = {f_col:>6.2E} Hz")


def _print_lengths(l_d, l_e, l_i, rho_e, rho_p, r_col):
    print("\nLengths: ")
    print("*" * 11)
    print(f"{'l_d':>5} = {l_d:>6.2E} m")
    print(f"{'d_e':>5} = {l_e:>6.2E} m")
    print(f"{'d_i':>5} = {l_i:>6.2E} m")
    print(f"{'r_e':>5} = {rho_e:>6.2E} m")
    print(f"{'r_p':>5} = {rho_p:>6.2E} m")
    print(f"{'r_col':>5} = {r_col:6.2E} m")


def _print_velocities(v_a, v_ae, v_te, v_tp, v_ts):
    print("\nVelocities: ")
    print("*" * 11)
    print(f"{'V_a':>5} = {v_a:>6.2E} m/s")
    print(f"{'V_ae':>5} = {v_ae:>6.2E} m/s")
    print(f"{'V_te':>5} = {v_te:>6.2E} m/s")
    print(f"{'V_tp':>5} = {v_tp:>6.2E} m/s")
    print(f"{'C_s':>5} = {v_ts:>6.2E} m/s")


def _print_other(n_d, eta, p_mag):
    print("\nOther parameters: ")
    print("*" * 17)
    print(
        f"{'N_deb':>5} = {n_d:>6.2E} {'':<6} " f"# number of particle in Debye sphere",
    )
    print(f"{'eta':>5} = {eta:>6.2E} {'Ohm m':<6} # Spitzer resistivity")
    print(f"{'P_B':>5} = {p_mag:>6.2E} {'Pa':<6} # Magnetic pressure")


def _print_dimensionless(beta, gamma_e):
    m_p = constants.proton_mass
    m_e = constants.electron_mass
    mp_me = m_p / m_e

    print("\nDimensionless parameters: ")
    print("*" * 25)
    print(f"{'beta':>20} = {beta:>6.2E} #  H+ beta")
    print(f"{'beta*sqrt(m_p/m_e)':>20} = {beta * np.sqrt(mp_me):>6.2E}")
    print(f"{'beta*(m_p/m_e)':>20} = {beta * mp_me:>6.2E}")
    print(f"{'gamma_e':>20} = {gamma_e:>6.2E} # 1/sqrt(1-(V_te/c)^2)")


def iplasma_calc(output: bool = False, verbose: bool = True):
    r"""Interactive function to calcute plasma paramters.

    Parameters
    ----------
    output : bool, Optional
        Flag to return dict with output. Default is False.
    verbose : bool, Optional
        Flag to print the results function. Default is True.

    Returns
    -------
    out : dict
        Hashtable with everything.

    """

    b_0 = float(input(f"{'Magnetic field in nT [10] ':<34}: ") or "10") * 1e-9
    n_hplus = float(input(f"{'H+ desity in cc [1] ':<34}: ") or "1") * 1e6
    t_e = float(
        input(f"{'Electron  temperature in eV [100] ':<34}: ") or "10",
    )
    t_i = float(input(f"{'Ion  temperature in eV [1000] ':<34}: ") or "1000")

    n_i, n_e = [n_hplus] * 2

    # Get constants
    q_e = constants.elementary_charge
    cel = constants.speed_of_light
    mu0 = constants.mu_0
    ep0 = constants.epsilon_0
    m_p = constants.proton_mass
    m_e = constants.electron_mass
    mp_me = m_p / m_e

    w_pe = np.sqrt(n_e * q_e**2 / (m_e * ep0))  # rad/s
    w_ce = q_e * b_0 / m_e  # rad/s
    w_pp = np.sqrt(n_i * q_e**2 / (m_p * ep0))

    p_mag = b_0**2 / (2 * mu0)

    v_a = b_0 / np.sqrt(mu0 * n_i * m_p)

    v_ae = b_0 / np.sqrt(mu0 * n_e * m_e)
    v_te = cel * np.sqrt(1 - 1 / (t_e * q_e / (m_e * cel**2) + 1) ** 2)
    v_tp = cel * np.sqrt(1 - 1 / (t_i * q_e / (m_p * cel**2) + 1) ** 2)
    # Sound speed formula (F. Chen, Springer 1984).
    v_ts = np.sqrt((t_e * q_e + 3 * t_i * q_e) / m_p)

    gamma_e = 1 / np.sqrt(1 - (v_te / cel) ** 2)
    gamma_p = 1 / np.sqrt(1 - (v_tp / cel) ** 2)

    l_e = cel / w_pe
    l_i = cel / w_pp
    # Debye length scale, sqrt(2) needed because of Vte definition
    l_d = v_te / (w_pe * np.sqrt(2))
    # number of e- in Debye sphere
    n_d = l_d * ep0 * m_e * v_te**2 / q_e**2

    f_pe = w_pe / (2 * np.pi)  # Hz
    f_ce = w_ce / (2 * np.pi)
    f_uh = np.sqrt(f_ce**2 + f_pe**2)
    f_pp = w_pp / (2 * np.pi)
    f_cp = f_ce / mp_me
    f_lh = np.sqrt(f_cp * f_ce / (1 + f_ce**2 / f_pe**2) + f_cp**2)

    rho_e = m_e * cel / (q_e * b_0) * np.sqrt(gamma_e**2 - 1)
    rho_p = m_p * cel / (q_e * b_0) * np.sqrt(gamma_p**2 - 1)
    rho_s = v_ts / (f_cp * 2 * np.pi)

    # Collision stuff
    # collision frequency e-/ions
    f_col = (n_e * q_e**4) / 16 * np.pi * ep0**2 * m_e**2 * v_te**3
    # Spitzer resistivity
    eta = np.pi * q_e**2 * np.sqrt(m_e)
    eta /= (4 * np.pi * ep0) ** 2 * (q_e * t_e) ** (3 / 2)
    eta *= np.log(4 * np.pi * n_d)
    # resistive scale
    r_col = eta / (mu0 * v_a)

    beta = v_tp**2 / v_a**2

    if verbose:
        _print_header()
        _print_frequencies(f_pe, f_ce, f_uh, f_lh, f_pp, f_cp, f_col)
        _print_lengths(l_d, l_e, l_i, rho_e, rho_p, r_col)
        _print_velocities(v_a, v_ae, v_te, v_tp, v_ts)
        _print_other(n_d, eta, p_mag)
        _print_dimensionless(beta, gamma_e)

    if output:
        out = {
            "w_pe": w_pe,
            "w_ce": w_ce,
            "w_pp": w_pp,
            "v_a": v_a,
            "v_ae": v_ae,
            "v_te": v_te,
            "v_tp": v_tp,
            "v_ts": v_ts,
            "gamma_e": gamma_e,
            "gamma_p": gamma_p,
            "l_e": l_e,
            "l_i": l_i,
            "l_d": l_d,
            "n_d": n_d,
            "f_pe": f_pe,
            "f_ce": f_ce,
            "f_uh": f_uh,
            "f_pp": f_pp,
            "f_cp": f_cp,
            "f_lh": f_lh,
            "rho_e": rho_e,
            "rho_p": rho_p,
            "rho_s": rho_s,
        }
    else:
        out = None

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _calc_diel(kc_, w_final, theta_, wp_e, wp_i, wc_i):
    # The elements of the dielectric tensor, using Swansons notation
    diel_s = 1 - wp_e**2 / (w_final**2 - 1) - wp_i**2 / (w_final**2 - wc_i**2)
    diel_d = -(wp_e**2) / (w_final * (w_final**2 - 1))
    diel_d += wc_i * wp_i**2 / (w_final * (w_final**2 - wc_i**2))
    diel_p = 1 - (wp_e**2 + wp_i**2) / w_final**2

    n2_ = kc_**2 / w_final**2

    diel_xx = diel_s - n2_ * np.cos(theta_) ** 2
    diel_xy = -1j * diel_d
    diel_xz = n2_ * np.cos(theta_) * np.sin(theta_)
    diel_yy = diel_s - n2_
    diel_zz = diel_p - n2_ * np.sin(theta_) ** 2
    return diel_xx, diel_xy, diel_xz, diel_yy, diel_zz


def _calc_e(diel_tensor):
    _, diel_xy, diel_xz, diel_yy, diel_zz = diel_tensor
    e_x = -diel_zz / diel_xz
    e_y = diel_xy / diel_yy * e_x
    e_z = np.ones(e_y.shape)

    e_per = np.sqrt(e_x * np.conj(e_x) + e_y * np.conj(e_y))
    e_pol = -2 * np.imag(e_x * np.conj(e_y)) / e_per**2
    e_tot = np.sqrt(e_x * np.conj(e_x) + e_y * np.conj(e_y) + e_z**2)

    return e_x, e_y, e_z, e_per, e_pol, e_tot


def _calc_b(kc_x_mat, kc_z_mat, w_final, e_x, e_y, e_z):
    b_x = -kc_z_mat * e_y / w_final
    b_y = (kc_z_mat * e_x - kc_x_mat * e_z) / w_final
    b_z = kc_x_mat * e_y / w_final

    b_par = np.sqrt(b_z * np.conj(b_z))
    b_per = np.sqrt(b_x * np.conj(b_x) + b_y * np.conj(b_y))
    b_pol = -2 * np.imag(b_x * np.conj(b_y)) / b_per**2
    b_tot = np.sqrt(
        b_x * np.conj(b_x) + b_y * np.conj(b_y) + b_z * np.conj(b_z),
    )

    return b_x, b_y, b_z, b_par, b_per, b_pol, b_tot


def _calc_s(e_x, e_y, e_z, b_x, b_y, b_z):
    # Poynting flux
    s_x = e_y * np.conj(b_z) - e_z * np.conj(b_y)
    s_y = e_z * np.conj(b_x) - e_x * np.conj(b_z)
    s_z = e_x * np.conj(b_y) - e_y * np.conj(b_x)
    s_par = np.abs(s_z)
    s_tot = np.sqrt(
        s_x * np.conj(s_x) + s_y * np.conj(s_y) + s_z * np.conj(s_z),
    )

    return s_par, s_tot


def _calc_part2fields(wp_e, en_e, en_i, e_tot, b_tot):
    n_e = wp_e**2
    en_e_n = en_e * n_e
    en_i_n = en_i * n_e
    en_efield = 0.5 * e_tot**2
    en_bfield = 0.5 * b_tot**2
    ratio_part_field = (en_e_n + en_i_n) / (en_efield + en_bfield)
    return ratio_part_field


def _calc_continuity(kc_x_mat, kc_z_mat, w_final, v_ex, v_ez, v_ix, v_iz):
    dn_e_n = (kc_x_mat * v_ex + kc_z_mat * v_ez) / w_final
    dn_e_n = np.sqrt(dn_e_n * np.conj(dn_e_n))
    dn_i_n = (kc_x_mat * v_ix + kc_z_mat * v_iz) / w_final
    dn_i_n = np.sqrt(dn_i_n * np.conj(dn_i_n))
    dne_dni = dn_e_n / dn_i_n

    return dn_e_n, dn_i_n, dne_dni


def _calc_vei(m_i, wc_i, w_final, e_x, e_y, e_z):
    q_e, q_i, m_e, wc_e = [-1, 1, 1, 1]

    v_ex = 1j * q_e * (w_final * e_x - 1j * wc_e * e_y)
    v_ex /= m_e * (w_final**2 - wc_e**2)

    v_ey = 1j * q_e * (1j * wc_e * e_x + w_final * e_y)
    v_ey /= m_e * (w_final**2 - wc_e**2)

    v_ez = 1j * q_e * e_z / (m_e * w_final)

    v_ix = 1j * q_i * (w_final * e_x + 1j * wc_i * e_y)
    v_ix /= m_i * (w_final**2 - wc_i**2)

    v_iy = 1j * q_i * (-1j * wc_i * e_x + w_final * e_y)
    v_iy /= m_i * (w_final**2 - wc_i**2)

    v_iz = 1j * q_i * e_z / (m_i * w_final)

    return v_ex, v_ey, v_ez, v_ix, v_iy, v_iz


def disp_surf_calc(kc_x_max, kc_z_max, m_i, wp_e):
    r"""Calculate the cold plasma dispersion surfaces according to equation
    2.64 in Plasma Waves by Swanson (2nd ed.)

    Parameters
    ----------
    kc_x_max : float
        Max value of k_perpendicular*c/w_c.
    kc_z_max : float
        Max value of k_parallel*c/w_c.
    m_i : float
        Ion mass in terms of electron masses.
    wp_e : float
        Electron plasma frequency in terms of electron gyro frequency.

    Returns
    -------
    kx_ : numpy.ndarray
        kperpandicular*c/w_c meshgrid
    kz_ : numpy.ndarray
        kparallel*c/w_c meshgrid
    wf_ : numpy.ndarray
        Dispersion surfaces.
    extra_param : dict
        Extra parameters to plot.
    """

    # Make vectors of the wave numbers
    kc_z = np.linspace(1e-6, kc_z_max, 35)
    kc_x = np.linspace(1e-6, kc_x_max, 35)

    # Turn those vectors into matrices
    kc_x_mat, kc_z_mat = np.meshgrid(kc_x, kc_z)

    # Find some of the numbers that appear later in the calculations
    kc_ = np.sqrt(kc_x_mat**2 + kc_z_mat**2)  # Absolute value of k
    theta_ = np.arctan2(kc_x_mat, kc_z_mat)  # The angle between k and B
    wc_i = 1 / m_i  # The ion gyro frequency
    wp_i = wp_e / np.sqrt(m_i)  # The ion plasma frequency
    wp_ = np.sqrt(wp_e**2 + wp_i**2)  # The total plasma frequency

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # For every k_perp and k_par, turn the dispersion relation into a
    # polynomial equation and solve it.
    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # The polynomial coefficients are calculated
    pol_koeff_8 = -2 * kc_**2
    pol_koeff_8 -= (1 + wc_i**2 + 3 * wp_**2) * np.ones(kc_.shape)
    pol_koeff_6 = (2 * kc_**2 + wp_**2) * (1 + wc_i**2 + 2 * wp_**2)
    pol_koeff_6 += kc_**4 + (wp_**2 + wc_i) ** 2
    pol_koeff_4 = -(kc_**4) * (1 + wc_i**2 + wp_**2)
    pol_koeff_4 -= 2 * kc_**2 * (wp_**2 + wc_i) ** 2
    pol_koeff_4 -= (kc_ * wp_) ** 2 * (1 + wc_i**2 - wc_i) * (1 + np.cos(theta_) ** 2)
    pol_koeff_4 -= wp_**2 * (wp_**2 + wc_i) ** 2
    pol_koeff_2 = kc_**4 * (
        wp_**2 * (1 + wc_i**2 - wc_i) * np.cos(theta_) ** 2 + wc_i * (wp_**2 + wc_i)
    )
    pol_koeff_2 += kc_**2 * wp_**2 * wc_i * (wp_**2 + wc_i) * (1 + np.cos(theta_) ** 2)
    pol_koeff_0 = -(kc_**4) * wc_i**2 * wp_**2 * np.cos(theta_) ** 2

    w_final = np.zeros((10, len(kc_z), len(kc_x)))

    # For each k, solve the equation
    for k_z, k_x in itertools.product(range(len(kc_z)), range(len(kc_x))):
        disp_polynomial = [
            1,
            0,
            pol_koeff_8[k_z, k_x],
            0,
            pol_koeff_6[k_z, k_x],
            0,
            pol_koeff_4[k_z, k_x],
            0,
            pol_koeff_2[k_z, k_x],
            0,
            pol_koeff_0[k_z, k_x],
        ]
        # theoretically should be real (A. Tjulin)
        w_temp = np.real(np.roots(disp_polynomial))
        # We need to sort the answers to get nice surfaces.
        w_final[:, k_z, k_x] = np.sort(w_temp)

    n2_ = kc_**2 / w_final**2
    v_ph_c = np.sqrt(1.0 / n2_)
    va_c = 1 / (wp_e * np.sqrt(m_i))
    v_ph_va = v_ph_c / va_c

    diel_tensor = _calc_diel(kc_, w_final, theta_, wp_e, wp_i, wc_i)

    e_x, e_y, e_z, _, e_pol, e_tot = _calc_e(diel_tensor)
    e_par = (kc_x_mat * e_x + kc_z_mat * e_z) / kc_

    b_x, b_y, b_z, b_par, _, b_pol, b_tot = _calc_b(
        kc_x_mat,
        kc_z_mat,
        w_final,
        e_x,
        e_y,
        e_z,
    )

    dk_x, dk_z = [kc_x_mat[1], kc_z_mat[1]]
    dw_x, dw_z = [np.zeros(w_final.shape) for _ in range(2)]
    dw_x[:, :, 1:] = np.diff(w_final, axis=2)
    dw_z[:, 1:, :] = np.diff(w_final, axis=1)
    v_x, v_z = [dw_ / dk for dw_, dk in zip([dw_x, dw_z], [dk_x, dk_z])]

    s_par, s_tot = _calc_s(e_x, e_y, e_z, b_x, b_y, b_z)

    # Compute ion and electron velocities
    v_ex, v_ey, v_ez, v_ix, v_iy, v_iz = _calc_vei(
        m_i,
        wc_i,
        w_final,
        e_x,
        e_y,
        e_z,
    )

    # Ratio of parallel and perpendicular to B speed
    vepar_perp = v_ez * np.conj(v_ez)
    vepar_perp /= v_ex * np.conj(v_ex) + v_ey * np.conj(v_ey)
    vipar_perp = v_iz * np.conj(v_iz)
    vipar_perp /= v_ix * np.conj(v_ix) + v_iy * np.conj(v_iy)

    # Total particle speeds
    v_e2 = v_ex * np.conj(v_ex) + v_ey * np.conj(v_ey) + v_ez * np.conj(v_ez)
    v_i2 = v_ix * np.conj(v_ix) + v_iy * np.conj(v_iy) + v_iz * np.conj(v_iz)

    # Ion and electron energies
    m_e = -1
    en_e = 0.5 * m_e * v_e2
    en_i = 0.5 * m_i * v_i2

    # Ratio of particle and field energy densities
    ratio_part_field = _calc_part2fields(wp_e, en_e, en_i, e_tot, b_tot)

    # Continuity equation
    dn_e_n, dn_i_n, dne_dni = _calc_continuity(
        kc_x_mat,
        kc_z_mat,
        w_final,
        v_ex,
        v_ez,
        v_ix,
        v_iz,
    )

    dn_e_n_db_b = dn_e_n / b_tot
    dn_i_n_db_b = dn_i_n / b_tot

    dn_e_n_dbpar_b = dn_e_n / b_par
    dn_i_n_dbpar_b = dn_i_n / b_par

    dn_e = dn_e_n * wp_e**2
    k_dot_e = e_x * kc_x_mat + e_z * kc_z_mat
    k_dot_e = np.sqrt(k_dot_e * np.conj(k_dot_e))

    # Build output dict
    extra_param = {
        "Degree of electromagnetism": np.log10(b_tot / e_tot),
        "Degree of longitudinality": np.abs(e_par) / e_tot,
        "Degree of parallelity E": e_z / e_tot,
        "Degree of parallelity B": np.sqrt(b_z * np.conj(b_z)) / b_tot,
        "W_E/W_B": np.log10(e_tot**2 / b_tot**2),
        "Ellipticity E": e_pol,
        "Ellipticity B": b_pol,
        "E_part/E_field": np.log10(ratio_part_field),
        "v_g": np.sqrt(v_x**2 + v_z**2),
        "v_ph/v_a": np.log10(v_ph_va),
        "E_e/E_i": np.log10(en_e / en_i),
        "v_e/v_i": np.log10(np.sqrt(v_e2 / v_i2)),
        "v_epara/v_eperp": np.log10(vepar_perp),
        "v_ipara/v_iperp": np.log10(vipar_perp),
        "dn_e/dn_i": np.log10(dne_dni),
        "(dn_e/n)/ (dB/B)": np.log10(dn_e_n_db_b),
        "(dn_i/n)/(dB/B)": np.log10(dn_i_n_db_b),
        "(dn_i/n)/(dBpar/B)": np.log10(dn_i_n_dbpar_b),
        "(dn_e/n)/(dB/B)": np.log10(dn_e / k_dot_e),
        "(dn_e/n)/(dBpar /B)": np.log10(dn_e_n_dbpar_b),
        " Spar/Stot": s_par / s_tot,
    }

    for k, v in zip(extra_param.keys(), extra_param.values()):
        extra_param[k] = np.transpose(np.real(v), [0, 2, 1])

    kx_ = np.transpose(kc_x_mat)
    kz_ = np.transpose(kc_z_mat)
    wf_ = np.transpose(w_final, [0, 2, 1])

    return kx_, kz_, wf_, extra_param



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
from numpy.typing import NDArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def datetime642iso8601(
    time: Union[np.datetime64, NDArray[np.datetime64]],
) -> NDArray[np.str_]:
    r"""Convert datetime64 in ns units to ISO 8601 time format .

    Parameters
    ----------
    time : numpy.datetime64 or numpy.ndarray
        Time in datetime64 in ns units.

    Returns
    -------
    time_iso8601 : numpy.ndarray
        Time in ISO 8601 format.

    See Also
    --------
    pyrfu.pyrf.datetime642iso8601

    """
    if isinstance(time, np.datetime64):
        time_datetime64 = np.atleast_1d(time).astype("datetime64[ns]")
    elif isinstance(time, np.ndarray) and time.dtype in [
        "<M8[ns]",
        "<M8[us]",
        "<M8[ms]",
    ]:
        time_datetime64 = time.astype("datetime64[ns]")
    else:
        raise TypeError("time must be numpy.datetime64 or numpy.ndarray")

    # Convert to string
    time_iso8601 = time_datetime64.astype(str)
    time_iso8601 = np.atleast_1d(np.squeeze(np.stack([time_iso8601])))

    return time_iso8601



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Any, Optional

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def movmean(inp: DataArray, window_size: Optional[int] = None) -> DataArray:
    r"""Computes running average of the inp over npts points.

    Parameters
    ----------
    inp : DataArray
        Time series of the input variable.
    window_size : int, Optional
        Number of points to average over. Default is a 4-point running average.

    Returns
    -------
    DataArray
        Time series of the input variable averaged over npts points.

    Raises
    ------
    TypeError
        If inp is not a DataArray.
    ValueError
        If window_size is smaller than 2 or larger than the length of the data.

    Notes
    -----
    Works also with 3D skymap distribution.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load ion pressure tensor

    >>> p_xyz_i = mms.get_data("Pi_gse_fpi_brst_l2", tint, mms_id)

    Running average the pressure tensor over 10s

    >>> fs = pyrf.calc_fs(p_xyz_i)
    >>> p_xyz_i = pyrf.movmean(p_xyz_i, int(10 * fs))

    """

    # Checks if input is a DataArray
    if not isinstance(inp, xr.DataArray):
        raise TypeError("Input must be a xarray.DataArray")

    # Gets input data and time
    time: NDArray[np.datetime64] = inp.time.data
    inp_data: NDArray[np.float32] = inp.data

    # Checks if window_size is defined
    if window_size is None:
        window_size = 2
    elif window_size < 2 or window_size > len(time):
        raise ValueError(
            "Window size must be larger than 2 and smaller than the length of the data."
        )

    # Checks if window_size is odd. If not, makes it odd.
    if window_size % 2:
        window_size -= 1

    # Preallocates output
    out_dat: NDArray[np.float32] = np.empty(
        (len(time) - window_size, *inp_data.shape[1:]), dtype=np.float32
    )

    # Computes moving average
    cum_sum: NDArray[np.float32] = np.cumsum(inp_data, axis=0)
    out_dat[...] = (
        cum_sum[window_size:, ...] - cum_sum[:-window_size, ...]
    ) / window_size

    # Gets coordinates
    coords: list[NDArray[Any]] = [
        time[int(window_size / 2) : -int(window_size / 2)],
        *[inp.coords[k].data for k in inp.dims[1:]],
    ]

    # Output in DataArray type
    out: DataArray = xr.DataArray(
        out_dat, coords=coords, dims=inp.dims, attrs=inp.attrs
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import random
import unittest

# 3rd party imports
import numpy as np
from ddt import data, ddt, unpack

# Local imports
from .. import models
from . import generate_timeline

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.4"
__status__ = "Prototype"


@ddt
class IgrfTestCase(unittest.TestCase):
    @data(
        (
            generate_timeline(64.0, 100, ref_time="1789-07-14T00:00:00.000000000"),
            "dipole",
        ),
    )
    @unpack
    def test_igrf_input_time(self, timeline, flag):
        with self.assertWarns(UserWarning):
            models.igrf(timeline.astype(np.int64) / 1e9, flag)

    @data((generate_timeline(64.0, 100), "bazinga!"))
    @unpack
    def test_igrf_input_flag(self, timeline, flag):
        with self.assertRaises(NotImplementedError):
            models.igrf(timeline.astype(np.int64) / 1e9, flag)

    @data((generate_timeline(64.0, 100), "dipole"))
    @unpack
    def test_igrf_output(self, timeline, flag):
        result = models.igrf(timeline.astype(np.int64) / 1e9, flag)
        self.assertIsInstance(result[0], np.ndarray)
        self.assertListEqual(list(result[0].shape), list(timeline.shape))
        self.assertIsInstance(result[1], np.ndarray)
        self.assertListEqual(list(result[1].shape), list(timeline.shape))


@ddt
class MagnetopauseNormalTestCase(unittest.TestCase):
    @data(
        (np.random.rand(3), random.randint(1, 10), random.randint(1, 10), "bazinga!!")
    )
    @unpack
    def test_magnetopause_normal_input(self, r_gsm, b_z_imf, p_sw, model):
        with self.assertRaises(NotImplementedError):
            models.magnetopause_normal(r_gsm, b_z_imf, p_sw, model)

    @data(
        (np.random.rand(3), random.randint(1, 10), random.randint(1, 10), "mp_shue97"),
        (np.random.rand(3), random.randint(1, 10), random.randint(1, 10), "bs97"),
        (np.random.rand(3), -random.randint(1, 10), random.randint(1, 10), "bs97"),
        (np.random.rand(3), random.randint(1, 10), random.randint(1, 10), "mp_shue98"),
        (np.random.rand(3), random.randint(1, 10), random.randint(1, 10), "bs98"),
    )
    @unpack
    def test_magnetopause_normal_output(self, r_gsm, b_z_imf, p_sw, model):
        models.magnetopause_normal(r_gsm, b_z_imf, p_sw, model)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools
from typing import List, Union

# 3rd party imports
import numpy as np
from matplotlib import ticker

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def annotate_heatmap(
    im,
    data: Union[np.ndarray, List[List[float]]] = None,
    valfmt: str = "{x:.2f}",
    textcolors: tuple = ("black", "white"),
    threshold: float = None,
    **textkw,
):
    r"""Annotate a heatmap.

    Parameters
    ----------
    im : matplotlib.image.AxesImage
        The AxesImage to be labeled.
    data : array_like, Optional
        Data used to annotate.  If None, the image's data is used.
    valfmt : str or matplotlib.ticker.Formatter, Optional
        The format of the annotations inside the heatmap..
    textcolors : dict, Optional
        A pair of colors.  The first is used for values below a threshold,
        the second for those above.
    threshold : float, Optional
        Value in data units according to which the colors from textcolors are
        applied.  If None (the default) uses the middle of the colormap as
        separation.
    **textkw
        All other arguments are forwarded to each call to `text` used to create
        the text labels.

    Returns
    -------
    texts : list
        Cells labels

    """

    if data is None:
        data = im.get_array()

    if not isinstance(data, (np.ndarray, list)):
        raise TypeError("data must be a numpy array")

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max()) / 2.0

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = {"horizontalalignment": "center", "verticalalignment": "center"}
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i, j in itertools.product(range(data.shape[0]), range(data.shape[1])):
        kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
        text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
        texts.append(text)

    return texts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib as mpl
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.14"
__status__ = "Prototype"

__all__ = ["plot_contour"]


def plot_contour(axis, inp, **kwargs):
    r"""Plot a contour plot.

    Parameters
    ----------
    axis : matplotlib.pyplot.subplotsaxes
        Target axis to plot.
    inp : xarray.DataArray
        Input 2D data to plot.

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments for matplotlib.pyplot.contour.

    Returns
    -------
    axis : matplotlib.pyplot.subplotsaxes
        Axis with contour plot.
    clines : matplotlib.contour.QuadContourSet
        Contour lines.

    """

    if not isinstance(axis, mpl.axes.Axes):
        raise TypeError("ax must be a matplotlib.pyplot.subplotsaxes.")

    if not isinstance(inp, xr.DataArray):
        raise TypeError("inp must be a xarray.DataArray.")

    # Get dimensions of the input data
    dims = inp.dims
    x, y = [inp[dim].data for dim in dims]

    # Plot contour
    clines = axis.contour(x, y, inp.data.T, **kwargs)

    return axis, clines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect
import datetime
import json
import os
import re
from typing import Mapping, Optional, Union

# 3rd party imports
import numpy as np
from dateutil import parser
from dateutil.rrule import DAILY, rrule

from pyrfu.mms.db_init import MMS_CFG_PATH

# Local imports
from pyrfu.pyrf.datetime642iso8601 import datetime642iso8601
from pyrfu.pyrf.iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def list_files(
    tint: list[str],
    mms_id: Union[str, int],
    var: Mapping[str, str],
    data_path: Optional[str] = "",
) -> list[str]:
    r"""Find available files in the data directories for `var`.

    Parameters
    ----------
    tint : list
        Time interval
    mms_id : str or int
        Index of the spacecraft
    var : dict
        Dictionary containing at least 4 keys
            * var["inst"] : name of the instrument
            * var["tmmode"] : data rate
            * var["lev"] : data level
            * var["dtype"] : data type
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.mms.mms_config.py`

    Returns
    -------
    file_names : list
        List of files corresponding to the parameters in the selected time
        interval

    """
    # Check path
    if not data_path:
        # Read the current version of the MMS configuration file
        with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
            config = json.load(fs)

        root_path = os.path.normpath(config["local"])
    else:
        root_path = os.path.normpath(data_path)

    # Make sure that the data path exists
    assert os.path.exists(root_path), f"{root_path} doesn't exist!!"

    # Check time interval
    if isinstance(tint, list):
        tint_array = np.array(tint)
    else:
        raise TypeError("tint must be a list!!")

    # Convert time interval to ISO 8601
    if isinstance(tint_array[0], str):
        tint_iso8601 = datetime642iso8601(iso86012datetime64(tint_array))
    else:
        raise TypeError("Values must be in str!!")

    files_out = []

    if not isinstance(mms_id, str):
        mms_id = str(mms_id)

    # directory and file name search patterns:
    # - assume directories are of the form:
    # (srvy, SITL): spacecraft/instrument/rate/level[/datatype]/year/month/
    # (brst): spacecraft/instrument/rate/level[/datatype]/year/month/day/
    # - assume file names are of the form:
    # spacecraft_instrument_rate_level[_datatype]_YYYYMMDD[hhmmss]_version.cdf

    file_name = (
        f"mms{mms_id}_{var['inst']}_{var['tmmode']}_{var['lev']}"
        + r"(_)?.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"
    )

    d_start = parser.parse(parser.parse(tint_iso8601[0]).strftime("%Y-%m-%d"))
    until_ = parser.parse(tint_iso8601[1]) - datetime.timedelta(seconds=1)
    days = rrule(DAILY, dtstart=d_start, until=until_)

    if var["dtype"] == "" or var["dtype"] is None:
        level_and_dtype = var["lev"]
    else:
        level_and_dtype = os.sep.join([var["lev"], var["dtype"]])

    for date in days:
        if var["tmmode"] == "brst":
            local_dir = os.sep.join(
                [
                    root_path,
                    f"mms{mms_id}",
                    var["inst"],
                    var["tmmode"],
                    level_and_dtype,
                    date.strftime("%Y"),
                    date.strftime("%m"),
                    date.strftime("%d"),
                ],
            )
        else:
            local_dir = os.sep.join(
                [
                    root_path,
                    f"mms{mms_id}",
                    var["inst"],
                    var["tmmode"],
                    level_and_dtype,
                    date.strftime("%Y"),
                    date.strftime("%m"),
                ],
            )

        if os.name == "nt":
            full_path = os.sep.join([re.escape(local_dir) + os.sep, file_name])
        else:
            full_path = os.sep.join([re.escape(local_dir), file_name])

        regex = re.compile(full_path)

        for root, _, files in os.walk(local_dir):
            for file in files:
                file_path = os.sep.join([root, file])

                matches = regex.match(file_path)
                if matches:
                    this_time = parser.parse(matches.groups()[1])
                    if d_start <= this_time <= until_:
                        this_file = {
                            "file_name": file,
                            "timetag": "",
                            "full_name": file_path,
                            "file_size": "",
                        }

                        if this_file not in files_out:
                            files_out.append(this_file)

    in_files = files_out

    file_name = r"mms.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    file_times = []

    regex = re.compile(file_name)

    for in_file in in_files:
        matches = regex.match(in_file["file_name"])
        if matches:
            file_times.append(
                (
                    in_file["file_name"],
                    parser.parse(matches.groups()[0]).timestamp(),
                    in_file["timetag"],
                    in_file["file_size"],
                ),
            )

    # sort in time
    sorted_files = sorted(file_times, key=lambda x: x[1])

    times = [t[1] for t in sorted_files]

    idx_min = bisect.bisect_left(times, parser.parse(tint_iso8601[0]).timestamp())

    # note: purposefully liberal here; include one extra file so that we
    # always get the burst mode
    # data
    if idx_min == 0:
        files_in_interval = []
        for sorted_file in sorted_files[idx_min:]:
            files_in_interval.append(
                {
                    "file_name": sorted_file[0],
                    "timetag": sorted_file[2],
                    "file_size": sorted_file[3],
                },
            )
    else:
        files_in_interval = []
        for sorted_file in sorted_files[idx_min - 1 :]:
            files_in_interval.append(
                {
                    "file_name": sorted_file[0],
                    "timetag": sorted_file[2],
                    "file_size": sorted_file[3],
                },
            )

    local_files = []

    file_names = [f["file_name"] for f in files_in_interval]

    for file_out in files_out:
        if file_out["file_name"] in file_names:
            local_files.append(file_out["full_name"])

    file_names = sorted(local_files)

    return file_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Optional, Union

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

# Local imports
from ..pyrf.calc_fs import calc_fs
from ..pyrf.cotrans import cotrans
from ..pyrf.resample import resample
from ..pyrf.ts_vec_xyz import ts_vec_xyz
from .dsl2gse import _transformation_matrix

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def dsl2gsm(
    inp: DataArray, defatt: Union[Dataset, np.ndarray], direction: Optional[int] = 1
) -> DataArray:
    r"""Transform time series from MMS's DSL to GSM.

    Parameters
    ----------
    inp : DataArray
        Input time series to convert.
    defatt : Dataset or numpy.ndarray
        Spacecraft attitude.
    direction : {1, -1}, Optional
        Direction of transformation. +1 DSL -> GSE, -1 GSE -> DSL.
        Default is 1.

    Returns
    -------
    DataArray
        Time series of the input field in the new coordinates systems.

    Raises
    ------
    TypeError
        If defatt is not xarray.Dataset or numpy.ndarray.

    Examples
    --------
    >>> from pyrfu.mms import get_data, load_ancillary, dsl2gse

    Define time interval

    >>> tint = ["2015-05-09T14:00:000", "2015-05-09T17:59:590"]

    Load magnetic field in spacecraft coordinates

    >>> b_xyz = get_data("b_dmpa_fgm_brst_l2", tint, 1)

    Load spacecraft attitude

    >>> defatt = load_ancillary("defatt", tint, 1)

    Transform magnetic field to GSE

    >>> b_gse = dsl2gse(b_xyz, defatt)

    """
    if isinstance(defatt, xr.Dataset):
        x = np.cos(np.deg2rad(defatt.z_dec)) * np.cos(
            np.deg2rad(defatt.z_ra.data),
        )
        y = np.cos(np.deg2rad(defatt.z_dec)) * np.sin(
            np.deg2rad(defatt.z_ra.data),
        )
        z = np.sin(np.deg2rad(defatt.z_dec))
        sax_gei = ts_vec_xyz(defatt.time.data, np.transpose(np.vstack([x, y, z])))
        sax_gsm = cotrans(sax_gei, "gei>gsm")
        spin_ax_gsm = resample(sax_gsm, inp, f_s=calc_fs(inp))
        spin_axis = spin_ax_gsm.data

    elif isinstance(defatt, np.ndarray) and len(defatt) == 3:
        spin_axis = np.atleast_2d(defatt)

    else:
        raise TypeError("DEFATT/SAX input must be xarray.Dataset or vector")

    # Compute transformation matrix
    transf_mat = _transformation_matrix(spin_axis, direction)

    out_data = np.einsum("kji,ki->kj", transf_mat, inp.data)

    out = inp.copy()
    out.data = out_data
    out.attrs["COORDINATE_SYSTEM"] = "GSM"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging
import os
from typing import Dict, Optional, Union

# 3rd party imports
import numba
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from scipy import fft
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

# Local imports
from pyrfu.pyrf.calc_fs import calc_fs

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


@numba.jit(nopython=True, fastmath=True)  # type: ignore
def _ww(
    s_ww: NDArray[np.complex128],
    scales_mat: NDArray[np.float64],
    sigma: float,
    frequencies_mat: NDArray[np.float64],
    f_nyq: float,
) -> NDArray[np.complex128]:
    # TODO : use nested for loop and math instead of numpy and test speed!!
    w_w: NDArray[np.complex128] = s_ww * np.exp(
        -sigma * sigma * ((scales_mat * frequencies_mat - f_nyq) ** 2) / 2,
    )
    w_w = w_w * np.sqrt(1.0)
    return w_w


@numba.jit(nopython=True, parallel=True, fastmath=True)  # type: ignore
def _power_r(
    power: NDArray[np.complex128], new_freq_mat: NDArray[np.float64]
) -> NDArray[np.float64]:
    power2: NDArray[np.float64] = np.absolute(
        (2 * np.pi) * np.conj(power) * power / new_freq_mat
    )
    return power2


@numba.jit(nopython=True, parallel=True, fastmath=True)  # type: ignore
def _power_c(
    power: NDArray[np.complex128], new_freq_mat: NDArray[np.float64]
) -> NDArray[np.complex128]:
    power2: NDArray[np.complex128] = (
        np.sqrt(np.absolute((2 * np.pi) / new_freq_mat)) * power
    )
    return power2


def wavelet(
    inp: DataArray,
    f_s: Optional[float] = None,
    f: Optional[list[float]] = None,
    n_freqs: Optional[int] = None,
    linear: Optional[Union[float, bool]] = None,
    wavelet_width: Optional[float] = None,
    cut_edge: Optional[bool] = True,
    return_power: Optional[bool] = True,
) -> Union[DataArray, Dataset]:
    """Computes wavelet spectrogram based on fast FFT algorithm.
    Parameters
    ----------
    inp : DataArray
        Input quantity.
    f_s : float, Optional
        Sampling frequency of the input time series.
    f : list, Optional
        Vector [f_min f_max], calculate spectra between frequencies
        f_min and f_max.
    n_freqs : int, Optional
        Number of frequency bins.
    linear : float or bool, Optional
        Linear spacing between frequencies of df.
    wavelet_width : float, Optional
        Width of the Morlet wavelet. Default 5.36.
    cut_edge : bool, Optional
        Set to True to set points affected by edge effects to NaN,
        False to keep edge affect points. Default True
    return_power : bool, Optional
        Set to True to return the power, False for complex wavelet
        transform. Default True.

    Returns
    -------
    DataArray or Dataset
        Wavelet transform of the input.


    Raises
    ------
    TypeError
        If linear keyword argument is not bool or float.
    ValueError
        If input is not 1D or 2D.

    """

    # Check input
    if not isinstance(inp, xr.DataArray):
        raise TypeError("Input must be a DataArray")

    if f_s is None:
        f_s = calc_fs(inp)

    if n_freqs is None:
        n_freqs = 200

    if wavelet_width is None:
        wavelet_width = 5.36

    if linear is not None:
        if isinstance(linear, float):
            delta_f: float = linear
            linear_df: bool = True
        elif isinstance(linear, bool) and linear:
            delta_f = 100.0
            linear_df = True
            logging.warning("Unknown input for linear delta_f set to 100")
        else:
            raise TypeError("linear keyword argument must be bool or float")
    else:
        delta_f = 100.0
        linear_df = False

    # Nyquist frequency and wavelet width
    f_nyq: float = f_s / 2
    sigma: float = wavelet_width / f_nyq

    # Frequency range
    if f is None:
        f_min: float = f_nyq / 10**2
        f_max: float = f_nyq / 10**-2
    else:
        f_min, f_max = sorted(f)

    if linear_df:
        scale_number: int = int(np.floor(f_nyq / delta_f))

        # Scales range
        scale_min: float = delta_f
        scale_max: float = scale_number * delta_f
        scales: NDArray[np.float64] = f_nyq / (
            np.linspace(scale_max, scale_min, scale_number, dtype=np.float64)
        )
    else:
        scale_number = n_freqs
        scale_min = np.log10(f_nyq / f_max)
        scale_max = np.log10(f_nyq / f_min)
        scales = np.logspace(scale_min, scale_max, scale_number, dtype=np.float64)

    # Unpack time and data.
    # Remove the last sample if the total number of samples is odd.
    if len(inp.time.data) % 2:
        time: NDArray[np.datetime64] = inp.time.data[:-1]
        data: NDArray[np.float64] = inp.data[:-1, ...].astype(np.float64)
    else:
        time = inp.time.data
        data = inp.data.astype(np.float64)

    # Preallocate power2
    if return_power:
        power2: NDArray[Union[np.float64, np.complex128]] = np.zeros(
            (len(time), n_freqs), dtype=np.float64
        )
    else:
        power2 = np.zeros((len(time), n_freqs), dtype=np.complex128)

    # Check for NaNs
    scales[np.isnan(scales)] = 0.0

    # Find the frequencies for an FFT of all data
    freq: NDArray[np.float64] = (
        f_nyq * np.arange(1, 1 + len(data) / 2) / (len(data) / 2)
    )

    # The frequencies corresponding to FFT
    freqs_fft: NDArray[np.float64] = np.hstack([0, freq, -np.flip(freq[:-1])])
    _, freqs_fft_mat = np.meshgrid(scales, freqs_fft, sparse=True)

    # Get the correct frequencies for the wavelet transform
    freqs_cwt: NDArray[np.float64] = f_nyq / scales
    freqs_cwt_mat, _ = np.meshgrid(freqs_cwt, freqs_fft, sparse=True)

    if data.ndim in [1, 2]:
        out_dict: Dict[str, object] = {}
    else:
        raise ValueError("Input data must be 1D or 2D")

    # if scalar add virtual axis
    if len(inp.shape) == 1:
        data = data[:, np.newaxis]

    # go through all the data columns
    for i in range(data.shape[1]):
        # Make the FFT of all data
        data_col: NDArray[np.float64] = data[:, i]

        # Wavelet transform of the data
        # Forward FFT
        s_w: NDArray[np.complex128] = fft.fft(data_col, workers=os.cpu_count())

        scales_mat, s_w_mat = np.meshgrid(scales, s_w, sparse=True)

        # Calculate the FFT of the wavelet transform
        w_w: NDArray[np.complex128] = _ww(
            s_w_mat, scales_mat, sigma, freqs_fft_mat, f_nyq
        )

        # Backward FFT
        power: NDArray[np.complex128] = fft.ifft(w_w, axis=0, workers=os.cpu_count())

        # Calculate the power spectrum
        if return_power:
            power2 = _power_r(power, np.tile(freqs_cwt_mat, (len(power), 1)))
        else:
            power2 = _power_c(power, np.tile(freqs_cwt_mat, (len(power), 1)))

        if cut_edge:
            censure = np.floor(2 * scales).astype(int)

            for j in range(scale_number):
                power2[: censure[j], j] = np.nan

                power2[len(data_col) - censure[j] : len(data_col), j] = np.nan

        if len(inp.shape) == 2:
            # Construct xarray.DataArray here
            out_dict[str(inp.comp.data[i])] = (
                ["time", "frequency"],
                np.fliplr(power2),
            )

    if len(inp.shape) == 1:
        out: Union[DataArray, Dataset] = xr.DataArray(
            np.fliplr(power2),
            coords=[time, np.flip(freqs_cwt)],
            dims=["time", "frequency"],
        )
    else:
        out = xr.Dataset(
            out_dict,
            coords={"time": time, "frequency": np.flip(freqs_cwt)},
        )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import scipy
import xarray as xr

__author__ = "Apostolos Kolokotronis"
__email__ = "apostolos.kolokotronis@irf.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def ts_convolve(inp, kernel, mode: str = "nearest"):
    r"""
    Compute the convolution of a time series of N-dimensional data with a N-dimensional
    kernel.
    ***Right now has only been tested for calculating moving averages of 1D time series
    ***

    The convolution is done using scipy.ndimage.convolve, with mode = "nearest"
    (read documentation for scipy.ndimage.convolve for more information).

    Parameters:
    ----------
    inp : xarray.DataArray
        The time series to be convolved with the kernel.

    kernel: nd.array
        The kernel to apply to inp for the convolution.

    Returns:
    -------
    out : xarray.DataArray
        An array containing the convolution of inp with kernel. Mode "valid" is applied
        from numpy.convolve, which does not affect the edges of the time series.
    """
    message = "Invalid input type. Input must be xarray.DataArary or xarray.Dataset"
    assert isinstance(inp, xr.DataArray), message

    if "time" not in inp.dims:
        message = "Invalid input dimensions. Input must have a 'time' dimension"
        raise ValueError(message)

    if inp.data.ndim <= 2 and inp.data.ndim > 0:
        if inp.data.ndim == 2:
            convolution = inp.copy()
            for comp in range(inp.data.shape[1]):
                convolution[:, comp] = ts_convolve(inp[:, comp], kernel, mode)

            out = convolution
        else:
            convolution = scipy.ndimage.convolve(
                input=inp.data, weights=kernel, mode=mode
            )
            convolution = xr.DataArray(
                convolution, coords=inp.coords, dims=inp.dims, attrs=inp.attrs
            )
            out = convolution
    else:

        raise ValueError("Invalid dimensions.")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pvi(inp, scale: int = 10):
    r"""Returns the PVI of a time series.

    .. math::

        y = \frac{|x_i - x_{i+s}|^2}{<|x_i - x_{i+s}|^2>}

    where :math:`s` is the scale.

    Parameters
    ----------
    inp : xarray.DataArray
        Input time series.
    scale : int, Optional
        Scale at which to compute the PVI. Default is 10.

    Returns
    -------
    values : xarray.DataArray
        An xarray containing the pvi of the original time series.

    """

    if len(inp.data.shape) == 1:
        data = inp.data[:, np.newaxis]
    else:
        data = inp.data

    delta_inp = np.abs((data[scale:, :] - data[:-scale, :]))
    delta_inp2 = np.sum(delta_inp**2, axis=1)
    sigma = np.mean(delta_inp2)
    result = np.array(delta_inp2 / sigma)

    time = inp.coords[inp.dims[0]].data

    result = xr.DataArray(
        result,
        coords=[time[0 : len(delta_inp)]],
        dims=[inp.dims[0]],
        attrs=inp.attrs,
    )

    result.attrs["units"] = "dimensionless"

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Optional

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def dynamic_press(
    n_s: DataArray, v_xyz: DataArray, specie: Optional[str] = "ions"
) -> DataArray:
    r"""Computes dynamic pressure.

    Parameters
    ----------
    n_s : DataArray
        Time series of the number density of the specie.
    v_xyz : DataArray
        Time series of the bulk velocity of the specie.
    specie : str, Optional
        Specie 'ions' or 'electrons'. Default 'ions'.

    Returns
    -------
    DataArray
        Time series of the dynamic pressure of the specie.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load ion bulk velocity and remove spintone

    >>> v_xyz_i = mms.get_data("vi_gse_fpi_fast_l2", tint, mms_id)
    >>> st_xyz_i = mms.get_data("sti_gse_fpi_fast_l2", tint, mms_id)
    >>> v_xyz_i = v_xyz_i - st_xyz_i

    Ion number density

    >>> n_i = mms.get_data("ni_fpi_fast_l2", tint, mms_id)

    Compute dynamic pressure

    >>> p = pyrf.dynamic_press(n_i, v_xyz_i, specie="ions")

    """

    # Check input
    if not isinstance(n_s, xr.DataArray):
        raise TypeError("n_s must be a xarray.DataArray")

    if not isinstance(v_xyz, xr.DataArray):
        raise TypeError("v_xyz must be a xarray.DataArray")

    if not isinstance(specie, str):
        raise TypeError("specie must be a string")

    # Check n_s and v_xyz shapes
    if n_s.ndim != 1:
        raise ValueError("n_s must be a scalar")

    if v_xyz.ndim != 2 or v_xyz.shape[1] != 3:
        raise ValueError("v_xyz must be a vector")

    # Check specie
    if specie.lower() not in ["ions", "electrons"]:
        raise ValueError("specie must be 'ions' or 'electrons'")

    if specie.lower() == "ions":
        mass = constants.proton_mass
    else:
        mass = constants.electron_mass

    # Get data
    n_s_data: np.ndarray = n_s.data
    v_xyz_data: np.ndarray = v_xyz.data

    # Compute dynamic pressure
    p_dyn: np.ndarray = n_s_data * mass * np.linalg.norm(v_xyz_data, axis=1) ** 2
    p_dyn_ts: DataArray = ts_scalar(n_s.time.data, p_dyn)

    return p_dyn_ts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from datetime import datetime

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def date_str(tint: list[str], fmt: int = 1) -> str:
    r"""Creates a string corresponding to time interval for output plot naming.

    Parameters
    ----------
    tint : list of str
        Time interval.
    fmt : int
        Format of the output :
            * 1 : "%Y%m%d_%H%M"
            * 2 : "%y%m%d%H%M%S"
            * 3 : "%Y%m%d_%H%M%S"_"%H%M%S"
            * 4 : "%Y%m%d_%H%M%S"_"%Y%m%d_%H%M%S"

    Returns
    -------
    out : str
        String corresponding to the time interval in the desired format.

    """

    # Check input
    assert isinstance(tint, list), "tint must be a list"
    assert isinstance(tint[0], str), "1st element of tint must be a string"
    assert isinstance(tint[1], str), "2nd element of tint must be a string"
    assert fmt in range(1, 5), "fmt must be 1, 2, 3, or 4"

    assert len(tint[0]) > 25, "tint[0] must be in %Y-%m-%dT%H:%M:%S.%f format"
    assert len(tint[1]) > 25, "tint[1] must be in %Y-%m-%dT%H:%M:%S.%f format"

    tint = [t_[:26] for t_ in tint]

    start_time = datetime.strptime(tint[0], "%Y-%m-%dT%H:%M:%S.%f")
    end_time = datetime.strptime(tint[1], "%Y-%m-%dT%H:%M:%S.%f")

    if fmt == 1:
        out = start_time.strftime("%Y%m%d_%H%M")
    elif fmt == 2:
        out = start_time.strftime("%y%m%d%H%M%S")
    elif fmt == 3:
        out = "_".join(
            [
                start_time.strftime("%Y%m%d_%H%M%S"),
                end_time.strftime("%H%M%S"),
            ],
        )
    else:
        out = "_".join(
            [
                start_time.strftime("%Y%m%d_%H%M%S"),
                end_time.strftime("%Y%m%d_%H%M%S"),
            ],
        )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np
from scipy.optimize import fminbound

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _magnetopause(theta, *args):
    r0, alpha, x0, y0 = args
    out = (
        r0**2 * (2.0 / (1 + np.cos(theta))) ** (2 * alpha)
        - 2
        * r0
        * (2.0 / (1 + np.cos(theta))) ** alpha
        * (x0 * np.cos(theta) + y0 * np.sin(theta))
        + x0**2
        + y0**2
    )

    return out


def _bow_shock(r_xy, *args):
    x, y = r_xy[:, 0], r_xy[:, 1]
    x0, y0 = args
    out = (x - x0) ** 2 + (y - y0) ** 2
    return out


def magnetopause_normal(
    r_gsm,
    b_z_imf,
    p_sw,
    model: str = "mp_shue1997",
    m_alfven: float = 4.0,
):
    r"""Computes the distance and normal vector to the magnetopause for
    [1]_ or [2]_ model. Or bow shock for [3]_ model.

    Parameters
    ----------
    r_gsm : array_like
        GSM position in Re.
    b_z_imf : float
        IMF Bz in nT.
    p_sw : float
        Solar wind dynamic pressure in nPa.
    model : {"mp_shue1997", "mp_shue1998", "bs97", "bs98"}, Optional
        Name of model :
            * 'mp_shue97'   : Shue et al., 1997 (Default)
            * 'mp_shue98'   : Shue et al., 1998
            * 'bs97'        : Bow shock, Farris & Russell 1994
            * 'bs98'        : Bow shock, Farris & Russell 1994
    m_alfven : float, Optional
        Alfvenic Mach number, only needed if bow shock model is used.

    Returns
    -------
    mindist : float
        Minimum distance to the magnetopause, in Re. Positive value if
        spacecraft is inside the magnetopause, negative if outside the
        magnetopause.
    n_vec : numpy.ndarray
        Normal vector to the magnetopause (pointing away from Earth).


    References
    ----------
    .. [1]  J.-H. Shue, J. K. Chao, H. C. Fu, C. T. Russell, P. Song,
            K. K. Khurana, and H. J. Singer (1997), A new functional form to
            study the solar wind control of the magnetopause size and shape,
            J. Geophys. Res., 102, 9497, doi: https://doi.org/10.1029/97JA00196

    .. [2]  J.-H. Shue, P. Song, C. T. Russell, J. T. Steinberg, J. K. Chao,
            G. Zastenker, O. L. Vaisber, S. Kokubun, H. J. Singer,
            T. R. Detman and H. Kawano (1998), Magnetopause location under
            extreme solar wind conditions, J. Geophys. Res., 103(A8), 17,
            691–17,700, doi: https://doi.org/10.1029/98JA01103

    .. [3]  M. H. Farris and C. T. Russell (1994), Determining the standoff
            distance of the bow shock: Mach number dependence and use of
            models, J. Geophys. Res., 99, 17681.
            doi: https://doi.org/10.1029/94JA01020

    """

    if model.lower() in ["mp_shue98", "bs98"]:
        logging.info("Shue et al., 1998 model used.")
        alpha = (0.58 - 0.007 * b_z_imf) * (1.0 + 0.024 * np.log(p_sw))
        r0 = 10.22 + 1.29 * np.tanh(0.184 * (b_z_imf + 8.14))
        r0 *= p_sw ** (-1.0 / 6.6)
    elif model.lower() in ["mp_shue97", "bs97"]:
        logging.info("Shue et al., 1997 model used.")
        alpha = (0.58 - 0.01 * b_z_imf) * (1.0 + 0.01 * p_sw)

        if b_z_imf >= 0:
            r0 = (11.4 + 0.013 * b_z_imf) * p_sw ** (-1.0 / 6.6)
        else:
            r0 = (11.4 + 0.140 * b_z_imf) * p_sw ** (-1.0 / 6.6)

    else:
        raise NotImplementedError(f"Invalid model : {model}")

    # Spacecraft position
    r1_x, r1_y, r1_z = r_gsm
    r0_x, r0_y = [r1_x, np.sqrt(r1_y**2 + r1_z**2)]

    if model[:2].lower() == "mp":
        # Magnetopause

        theta_min, min_val, _, _ = fminbound(
            _magnetopause,
            x1=-np.pi / 2,
            x2=np.pi / 2,
            args=(r0, alpha, r0_x, r1_y),
            full_output=True,
        )

        min_dist = np.sqrt(min_val)

        # calculate the direction to the spacecraft normal to the magnetopause
        x_n = r0 * (2 / (1 + np.cos(theta_min))) ** alpha * np.cos(theta_min) - r1_x
        phi = np.arctan2(r1_z, r1_y)
        y_n = (
            np.cos(phi)
            * (r0 * (2 / (1 + np.cos(theta_min))) ** alpha * np.sin(theta_min))
            - r1_y
        )
        z_n = (
            np.sin(phi)
            * (r0 * (2 / (1 + np.cos(theta_min))) ** alpha * np.sin(theta_min))
            - r1_z
        )

        n_vec = np.stack([x_n, y_n, z_n]) / min_dist

        # if statement to ensure normal is pointing away from Earth
        fact = r0 * (2 / (1 + np.cos(theta_min))) ** alpha - np.sqrt(r0_x**2 + r0_y**2)
        n_vec *= np.sign(fact)
        min_dist *= np.sign(fact)

    else:
        # Bow shock
        logging.info("Farris & Russell 1994 bow shock model used.")

        gamma = 5 / 3
        mach = m_alfven

        # Bow shock standoff distance
        rbs = r0 * (
            1 + 1.1 * ((gamma - 1) * mach**2 + 2) / ((gamma + 1) * (mach**2 - 1))
        )

        # y ^ 2 = 0 - Ax + Bx ^ 2
        coeffs = [0, 45.3, 0.04]
        x = np.linspace(rbs, -100, int((rbs + 100) * 1e3) + 1)
        y = np.sqrt(-coeffs[1] * (x - rbs) + coeffs[2] * (x - rbs) ** 2)
        x = np.hstack([np.flip(x), x])
        y = np.hstack([-np.flip(y), y])

        r_xy = np.transpose(np.stack([x, y]))
        args_bow_shock = (r0_x, r0_y)

        min_val = np.min(_bow_shock(r_xy, *args_bow_shock))
        min_pos = np.argmin(_bow_shock(r_xy, *args_bow_shock))

        d_vec = np.transpose(np.vstack([x - r0_x, y - r0_y]))
        d_min = d_vec[min_pos, :]

        x_n = d_min[0] / np.linalg.norm(d_min)
        min_dist = np.sqrt(min_val)

        qyz = r1_y / r1_z
        z_n = np.sign(r1_z) * np.sign(x_n) * np.sqrt((1 - x_n**2) / (1 + qyz**2))
        y_n = z_n * qyz

        n_vec = np.stack([x_n, y_n, z_n])

        # if statement to ensure normal is pointing away from Earth
        min_dist *= np.sign(n_vec[0])
        n_vec *= np.sign(n_vec[0])

    return min_dist, n_vec



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from matplotlib.axes import Axes
from matplotlib.dates import num2date
from xarray.core.dataarray import DataArray

# Local imports
from ..pyrf.t_eval import t_eval

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def add_position(
    ax: Axes,
    r_xyz: DataArray,
    spine: float = 20,
    position: str = "top",
    fontsize: float = 10,
) -> Axes:
    r"""Add extra axes to plot spacecraft position.

    Parameters
    ----------
    ax : matplotlib.axes._axes.Axes
        Axis where to label the spacecraft position.
    r_xyz : xarray.DataArray
        Time series of the spacecraft position.
    spine : float, Optional
        Relative position of the axes. Default is 20.
    position : str, Optional
        Axis position wtr to the reference axis. Default is "top".
    fontsize : float, Optional
        xticks label font size. Default is 10.

    Returns
    -------
    axr : matplotlib.axes._axes.Axes
        Twin axis with spacecraft position as x-axis label.

    """

    x_lim = ax.get_xlim()

    t_ticks = [t_.replace(tzinfo=None) for t_ in num2date(ax.get_xticks())]
    t_ticks = np.array(t_ticks).astype("<M8[ns]")
    r_ticks = t_eval(r_xyz, t_ticks).data

    ticks_labels = []
    for ticks_ in r_ticks:
        ticks_labels.append(
            f"{ticks_[0]:3.2f}\n{ticks_[1]:3.2f}\n{ticks_[2]:3.2f}",
        )

    axr = ax.twiny()
    axr.spines[position].set_position(("outward", spine))
    axr.xaxis.set_ticks_position(position)
    axr.xaxis.set_label_position(position)
    axr.set_xticks(t_ticks)
    axr.set_xticklabels(ticks_labels, fontsize=fontsize)
    axr.set_xlim(x_lim)

    return axr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf.calc_fs import calc_fs
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def dft_time_shift(inp, tau):
    r"""Shifts the input signal ``inp`` by ``tau`` seconds using discrete
    fourier transform (DFT). Particularly useful when calculating the
    frequency-wavenumber spectrum of the mms' spin-plane or axial probes.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to be shifted (Note : Only tensor order 1).
    tau : float
        Applied shift in seconds.

    Returns
    -------
    out : xarray.DataArray
        Time series of the shifted input.

    See also
    --------
    pyrfu.mms.fk_power_spectrum : Calculates the frequency-wave number
    power spectrum.

    """

    time, sig = [inp.time.data, inp.data]

    # Sampling frequency
    f_sampling = calc_fs(inp)

    # Applied delay in samples.
    delay = np.floor(tau * f_sampling)

    # Forward FFT
    sig_fft = np.fft.fft(sig)
    n_p = len(sig)

    # Disregard Nyquist frequency for even-sized dft
    if not n_p % 2:
        sig_fft[int(n_p / 2 + 1)] = 0

    freq = (np.arange(n_p) + np.floor(n_p / 2)) % n_p - np.floor(n_p / 2)
    freq /= n_p

    # Backward FFT
    out_data = np.fft.ifft(sig_fft * np.exp(-2j * np.pi * delay * freq))

    out_time = time + int(tau * 1e9)

    out = ts_scalar(out_time, out_data, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import List

# 3rd party imports
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pl_tx(
    axis: Axes, inp_list: List[DataArray], comp: int = 0, colors: str = "mms", **kwargs
) -> Axes:
    r"""Line plot of 4 spacecraft time series.

    Parameters
    ----------
    axis : matplotlib.axes._axes.Axes
        Axis
    inp_list : list of xarray.DataArray
        Time series to plot
    comp: int, Optional
        Index of the column to plot. Default is 0.
    colors: {'cluster', 'mms'}, Optional
        Color cycle to use. Default uses MMS

    Other Parameters
    ----------------
    kwargs : dict
        Hash table of plot options.

    Returns
    -------
    axis : matplotlib.axes._axes.Axes
        Axis with matplotlib.lines.Line2D.

    Raises
    ------
    NotImplementedError: if invalid color style of inp_list.ndim > 3

    """

    if axis is None:
        _, axis = plt.subplots(1)

    if colors.lower() not in ["cluster", "mms"]:
        raise NotImplementedError("Unknown color cycle")

    for i, inp in enumerate(inp_list):
        if inp.ndim == 1:
            data = inp.data[:, np.newaxis]
        elif inp.ndim == 2:
            data = inp.data
        elif inp.ndim == 3:
            data = np.reshape(
                inp.data,
                (inp.shape[0], inp.shape[1] * inp.shape[2]),
            )
        else:
            raise NotImplementedError("inp.ndim > 3 not implemented")

        time = inp.time
        axis.plot(time, data[:, comp], color=f"{colors}:{colors}{i + 1:d}", **kwargs)

    locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
    formatter = mdates.ConciseDateFormatter(locator)
    axis.xaxis.set_major_locator(locator)
    axis.xaxis.set_major_formatter(formatter)
    axis.grid(True, which="major", linestyle="-", linewidth="0.5", c="0.5")

    return axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from pycdfpp import _pycdfpp, load, to_datetime64
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def _pycdfpp_attributes_to_dict(attributes):
    attributes_dict = {}

    for k in attributes:
        tmp = [attributes[k][i] for i in range(len(attributes[k]))]

        if np.size(tmp) == 1:
            if isinstance(tmp[0], (list, np.ndarray)) and isinstance(
                tmp[0][0], _pycdfpp.tt2000_t
            ):
                attributes_dict[k] = to_datetime64(tmp[0][0])
            else:
                attributes_dict[k] = tmp[0]
        else:
            attributes_dict[k] = tmp[:]

    return attributes_dict


def get_variable(file_path: str, cdf_name: str) -> DataArray:
    r"""Read field named cdf_name in file and convert to DataArray.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.
    cdf_name : str
        Name of the target variable in the cdf file.

    Returns
    -------
    out : DataArray
        Target variable.

    """
    # Load file
    file = load(file_path)

    var_data = file[cdf_name].values
    var_attributes = _pycdfpp_attributes_to_dict(file[cdf_name].attributes)

    out = xr.DataArray(
        var_data,
        coords=[np.arange(len(var_data))],
        dims=["x"],
        attrs=var_attributes,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def cart2sph(x, y, z):
    r"""Cartesian to spherical coordinate transform.

    .. math::

        \alpha = \arctan \left( \frac{y}{x} \right) \\
        \beta = \arccos \left( \frac{z}{r} \right) \\
        r = \sqrt{x^2 + y^2 + z^2}

    with :math:`\alpha \in [0, 2\pi], \beta \in [0, \pi], r \geq 0`

    Parameters
    ----------
    x : float or array_like
        x-component of Cartesian coordinates
    y : float or array_like
        y-component of Cartesian coordinates
    z : float or array_like
        z-component of Cartesian coordinates

    Returns
    -------
    alpha : float or array_like
        Azimuth angle in radians
    beta : float or array_like
        Elevation angle in radians (with 0 denoting North pole)
    r : float or array_like
        Radius
    """

    # Radius
    r = np.sqrt(x**2 + y**2 + z**2)

    # Azimuthal angle
    alpha = np.arctan2(y, x)

    # Elevation angle
    beta = np.arccos(z / r)

    return alpha, beta, r



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import random
from math import asin, cos, sin, sqrt

# Third party imports
import numba
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def int_sph_dist(vdf, velocity, phi, theta, velocity_grid, phi_grid, **kwargs):
    r"""Integrate a spherical distribution function to a line/plane.

    Parameters
    ----------
    vdf : numpy.ndarray
        Phase-space density skymap.
    velocity : numpy.ndarray
        Velocity of the instrument bins,
    phi : numpy.ndarray
        Azimuthal angle of the instrument bins.
    theta : numpy.ndarray
        Elevation angle of the instrument bins.
    velocity_grid : numpy.ndarray
        Velocity grid for interpolation.
    phi_grid : numpy.ndarray
        Azimuthal angle grid for interpolation.
    **kwargs
        Keyword arguments.

    Returns
    -------
    pst : dict
        Dictionary with the projected distribution and
        corresponding velocity grid information.

    """

    # Coordinates system transformation matrix
    xyz = kwargs.get("xyz", np.eye(3))

    # Make sure the transformation matrix is orthonormal.
    x_phat = xyz[:, 0] / np.linalg.norm(xyz[:, 0])  # re-normalize
    y_phat = xyz[:, 1] / np.linalg.norm(xyz[:, 1])  # re-normalize

    z_phat = np.cross(x_phat, y_phat)
    z_phat /= np.linalg.norm(z_phat)
    y_phat = np.cross(z_phat, x_phat)

    # Define the rotation matrix from original to primed frame
    r_mat = np.transpose(np.stack([x_phat, y_phat, z_phat]), [1, 0])

    # Number of Monte Carlo iterations and how number of MC points is
    # weighted to data.
    n_mc = kwargs.get("n_mc", 10)
    weight = kwargs.get("weight", None)

    # limit on out-of-plane velocity
    v_lim = np.array(kwargs.get("v_int", [-np.inf, np.inf]), dtype=np.float64)

    # limit on azymuthal angle from projection plane
    a_lim = np.array(kwargs.get("a_int", [-180.0, 180.0]), dtype=np.float64)
    a_lim = np.deg2rad(a_lim)

    # Projection dimension and base
    projection_base = kwargs.get("projection_base", "pol")
    projection_dim = kwargs.get("projection_dim", "1d")

    velocity_edges = kwargs.get("velocity_edges", None)
    velocity_grid_edges = kwargs.get("velocity_grid_edges", None)

    # Azimuthal and elevation angles steps. Assumed to be constant
    # if not provided.
    d_phi = np.abs(np.median(np.diff(phi))) * np.ones_like(phi)
    d_phi = kwargs.get("d_phi", d_phi)
    d_theta = np.abs(np.median(np.diff(theta))) * np.ones_like(theta)
    d_theta = kwargs.get("d_theta", d_theta)

    if velocity_edges is None:
        d_v = np.hstack([np.diff(velocity[:2]), np.diff(velocity)])
        d_v_m, d_v_p = [np.diff(velocity) / 2.0] * 2
    else:
        d_v_m = velocity - velocity_edges[:-1]
        d_v_p = velocity_edges[1:] - velocity
        d_v = d_v_m + d_v_p

    # Overwrite projection dimension if azimuthal angle of projection
    # plane is not provided. Set the azimuthal angle grid width.
    if phi_grid is not None and projection_dim.lower() in ["2d", "3d"]:
        d_phi_grid = np.median(np.diff(phi_grid))
    else:
        projection_dim = "1d"
        d_phi_grid = 1.0

    # Speed grid bins edges
    if velocity_grid_edges is None:
        velocity_grid_diff = np.diff(velocity_grid)
        velocity_grid_edges = np.zeros(len(velocity_grid) + 1)
        velocity_grid_edges[0] = velocity_grid[0] - velocity_grid_diff[0] / 2.0
        velocity_grid_edges[1:-1] = velocity_grid[:-1] + velocity_grid_diff / 2.0
        velocity_grid_edges[-1] = velocity_grid[-1] + velocity_grid_diff[-1] / 2.0
    else:
        velocity_grid = velocity_grid_edges[:-1] + np.diff(velocity_grid_edges) / 2.0

    if projection_base == "pol":
        d_v_grid = np.diff(velocity_grid_edges)

        if projection_dim == "2d":
            raise NotImplementedError(
                "2d projection on polar grid is not ready yet!!",
            )
    else:
        mean_diff = np.mean(np.diff(velocity_grid))
        msg = "For a cartesian grid, all velocity bins must be equal!!"
        assert (np.diff(velocity_grid) / mean_diff - 1 < 1e-2).all(), msg

        d_v_grid = mean_diff

    # Weighting of number of Monte Carlo particles
    n_sum = n_mc * np.sum(vdf != 0)  # total number of Monte Carlo particles
    if weight == "lin":
        n_mc_mat = np.ceil(n_sum / np.sum(vdf) * vdf)
    elif weight == "log":
        n_mc_mat = np.ceil(
            n_sum / np.sum(np.log10(vdf + 1.0)) * np.log10(vdf + 1.0),
        )
    else:
        n_mc_mat = np.zeros_like(vdf)
        n_mc_mat[vdf != 0] = n_mc

    n_mc_mat = n_mc_mat.astype(int)

    if projection_base == "pol":
        d_a_grid = velocity_grid ** (int(projection_dim[0]) - 1) * d_phi_grid * d_v_grid
        d_a_grid = d_a_grid.astype(np.float64)
    else:
        d_a_grid = d_v_grid ** int(projection_dim[0])
        d_a_grid = d_a_grid.astype(np.float64)

    if projection_base == "cart" and projection_dim == "2d":
        f_g = _mc_cart_2d(
            vdf,
            velocity,
            phi,
            theta,
            d_v,
            d_v_m,
            d_phi,
            d_theta,
            velocity_grid_edges,
            d_a_grid,
            v_lim,
            a_lim,
            n_mc_mat,
            r_mat,
        )
    elif projection_base == "cart" and projection_dim == "3d":
        f_g = _mc_cart_3d(
            vdf,
            velocity,
            phi,
            theta,
            d_v,
            d_v_m,
            d_phi,
            d_theta,
            velocity_grid_edges,
            d_a_grid,
            v_lim,
            a_lim,
            n_mc_mat,
            r_mat,
        )
    elif projection_base == "pol" and projection_dim == "1d":
        f_g = _mc_pol_1d(
            vdf,
            velocity,
            phi,
            theta,
            d_v,
            d_v_m,
            d_phi,
            d_theta,
            velocity_grid_edges,
            d_a_grid,
            v_lim,
            a_lim,
            n_mc_mat,
            r_mat,
        )
    else:
        raise NotImplementedError(
            f"{projection_dim} projection on {projection_base} grid is not ready yet!!",
        )

    if projection_dim == "2d" and projection_base == "cart":
        pst = {
            "f": f_g,
            "vx": velocity_grid,
            "vy": velocity_grid,
            "vx_edges": velocity_grid_edges,
            "vy_edges": velocity_grid_edges,
        }
    elif projection_dim == "3d" and projection_base == "cart":
        pst = {
            "f": f_g,
            "vx": velocity_grid,
            "vy": velocity_grid,
            "vz": velocity_grid,
            "vx_edges": velocity_grid_edges,
            "vy_edges": velocity_grid_edges,
            "vz_edges": velocity_grid_edges,
        }
    else:
        pst = {"f": f_g, "vx": velocity_grid, "vx_edges": velocity_grid_edges}

    return pst


@numba.jit(cache=True, nogil=True, parallel=True, nopython=True)
def _mc_pol_1d(
    vdf,
    v,
    phi,
    theta,
    d_v,
    d_v_m,
    d_phi,
    d_theta,
    vg_edges,
    d_a_grid,
    v_lim,
    a_lim,
    n_mc,
    r_mat,
):
    r"""Perform 3D Monte-Carlo interpolation of the VDFs

    Parameters
    ----------
    vdf : double
        3D skymap particle velocity distribution function.
    v : double
        1D array of instrument velocity bins centers.
    phi : double
        1D array of instrument azimuthal angles bins centers.
    theta : double
        1D array of instrument elevation angles bins centers.
    d_v : double
        1D array of instrument velocity bins widths.
    d_v_m : double
        1D array of minus velocity from bins centers.
    d_phi : double
        1D array of instrument azimuthal angles bins widths.
    d_theta : double
        1D array of instrument elevation angles bins widths.
    vg_egdes : double
        Bin centers of the velocity of the projection grid.
    d_a_grid : double
        Bin centers of the azimuthal angle of the projection in radians in
        the span [0,2*pi]. If this input is given, the projection will be 2D.
        If it is omitted, the projection will be 1D.
    v_lim : double
        Limits on the out-of-plane velocity interval in 2D and "transverse"
        velocity in 1D.
    a_lim : double
        Angular limit in degrees, can be combined with v_lim.
    n_mc : double
        Number of Monte-Carlo particle for the corresponding instrument bins.
    r_mat : double
        Frame transformation matrix.

    Returns
    -------
    f_g : double
        Reduced/interpolated distribution.

    """

    n_v, n_ph, n_th = vdf.shape
    n_vg = len(vg_edges) - 1
    f_g = np.zeros(n_vg)

    for i in numba.prange(n_v):
        for j in range(n_ph):
            for k in range(n_th):
                n_mc_ijk = n_mc[i, j, k]

                if vdf[i][j][k] == 0.0:
                    continue

                dtau_ijk = v[i] ** 2 * np.cos(theta[k]) * d_v[i] * d_phi[j] * d_theta[k]
                c_ijk = dtau_ijk / n_mc_ijk
                f_ijk = vdf[i, j, k]

                for l_mc in range(n_mc_ijk):
                    # Generate Monte-Carlo particle speed, phi, and theta
                    if l_mc == 0:
                        # First Monte-Carlo particle is set at the bin center
                        d_v_mc = 0.0
                        d_phi_mc = 0.0
                    else:
                        d_v_mc = -random.random() * d_v[i] - d_v_m[0]
                        d_phi_mc = (random.random() - 0.5) * d_phi[j]

                    # convert instrument bin to cartesian velocity
                    v_mc = v[i] + d_v_mc
                    phi_mc = phi[j] + d_phi_mc

                    # This is needed to distribute the points evenly in space
                    # (more points further 'equatorward' since the slice is
                    # wider there).

                    if l_mc == 0:
                        theta_mc = theta[k]
                    else:
                        theta_1 = theta[k] - 0.5 * d_theta[k]
                        theta_2 = theta[k] + 0.5 * d_theta[k]
                        sin_theta_1 = sin(theta_1)
                        sin_theta_2 = sin(theta_2)
                        d_sin_theta = sin_theta_2 - sin_theta_1
                        sin_theta_mc = sin_theta_1 + random.random() * d_sin_theta
                        theta_mc = asin(sin_theta_mc)

                    v_x = v_mc * cos(theta_mc) * cos(phi_mc)
                    v_y = v_mc * cos(theta_mc) * sin(phi_mc)
                    v_z = v_mc * sin(theta_mc)

                    # Get velocities in primed coordinate system
                    # vxp = [vx, vy, vz] * xphat'; % all MC points
                    v_x_p = r_mat[0, 0] * v_x + r_mat[1, 0] * v_y + r_mat[2, 0] * v_z
                    v_y_p = r_mat[0, 1] * v_x + r_mat[1, 1] * v_y + r_mat[2, 1] * v_z
                    v_z_p = r_mat[0, 2] * v_x + r_mat[1, 2] * v_y + r_mat[2, 2] * v_z

                    # get transverse velocity sqrt(vy^2+vz^2)
                    v_z_p = sqrt(v_y_p**2 + v_z_p**2)
                    alpha = asin(v_z_p / v_mc)

                    # Check if the velocity is within the v_lim and a_lim bounds
                    use_point = v_lim[0] <= v_z_p < v_lim[1]
                    use_point = use_point and (a_lim[0] <= alpha < a_lim[1])

                    # Find which bin the MC point falls into
                    v_p = v_x_p

                    if v_p > vg_edges[-1] or v_p < vg_edges[0]:
                        continue

                    # searchsorted returns the index where the element should be
                    # inserted to maintain order, so we subtract 1 to get the index of
                    # the bin that contains the element
                    i_vxg = np.searchsorted(vg_edges, v_p) - 1

                    # Special case for the first bin edge, which is the lower limit of
                    # the first bin and should be included in the first bin
                    if i_vxg == -1:
                        i_vxg = 0

                    d_a = d_a_grid[i_vxg]

                    if use_point * (i_vxg < n_vg):
                        f_g[i_vxg] += f_ijk * c_ijk / d_a

    return f_g


@numba.jit(cache=True, nogil=True, parallel=True, nopython=True)
def _mc_cart_2d(
    vdf,
    v,
    phi,
    theta,
    d_v,
    d_v_m,
    d_phi,
    d_theta,
    vg_edges,
    d_a_grid,
    v_lim,
    a_lim,
    n_mc,
    r_mat,
):
    r"""Perform 3D Monte-Carlo interpolation of the VDFs

    Parameters
    ----------
    vdf : double
        3D skymap particle velocity distribution function.
    v : double
        1D array of instrument velocity bins centers.
    phi : double
        1D array of instrument azimuthal angles bins centers.
    theta : double
        1D array of instrument elevation angles bins centers.
    d_v : double
        1D array of instrument velocity bins widths.
    d_v_m : double
        1D array of minus velocity from bins centers.
    d_phi : double
        1D array of instrument azimuthal angles bins widths.
    d_theta : double
        1D array of instrument elevation angles bins widths.
    vg_egdes : double
        Bin centers of the velocity of the projection grid.
    d_a_grid : double
        Bin centers of the azimuthal angle of the projection in radians in
        the span [0,2*pi]. If this input is given, the projection will be 2D.
        If it is omitted, the projection will be 1D.
    v_lim : double
        Limits on the out-of-plane velocity interval in 2D and "transverse"
        velocity
        in 1D.
    a_lim : double
        Angular limit in degrees, can be combined with v_lim.
    n_mc : double
        Number of Monte-Carlo particle for the corresponding instrument bins.
    r_mat : double
        Frame transformation matrix.

    Returns
    -------
    f_g : double
        Reduced/interpolated distribution.

    """

    # Get dimension of the instrument and interpolation grid.
    n_v, n_ph, n_th = vdf.shape
    n_vg = len(vg_edges) - 1
    f_g = np.zeros((n_vg, n_vg))

    for i in range(n_v):
        for j in range(n_ph):
            for k in range(n_th):
                n_mc_ijk = n_mc[i, j, k]

                if vdf[i][j][k] == 0.0:
                    continue

                dtau_ijk = v[i] ** 2 * cos(theta[k]) * d_v[i] * d_phi[j] * d_theta[k]
                c_ijk = dtau_ijk / n_mc_ijk
                f_ijk = vdf[i, j, k]

                for l_mc in range(n_mc_ijk):
                    # Generate Monte-Carlo particle speed, phi, and theta
                    if l_mc == 0:
                        # First Monte-Carlo particle is set at the bin center
                        d_v_mc = 0.0
                        d_phi_mc = 0.0
                    else:
                        d_v_mc = -random.random() * d_v[i] - d_v_m[0]
                        d_phi_mc = (random.random() - 0.5) * d_phi[j]

                    # convert instrument bin to cartesian velocity
                    v_mc = v[i] + d_v_mc
                    phi_mc = phi[j] + d_phi_mc

                    # This is needed to distribute the points evenly in space
                    # (more points further 'equatorward' since the slice is
                    # wider there).

                    if l_mc == 0:
                        theta_mc = theta[k]
                    else:
                        theta_1 = theta[k] - 0.5 * d_theta[k]
                        theta_2 = theta[k] + 0.5 * d_theta[k]
                        sin_theta_1 = sin(theta_1)
                        sin_theta_2 = sin(theta_2)
                        d_sin_theta = sin_theta_2 - sin_theta_1
                        sin_theta_mc = sin_theta_1 + random.random() * d_sin_theta
                        theta_mc = asin(sin_theta_mc)

                    # Calculate velocity of the Monte-Carlo particle in
                    # cartesian coordinates
                    v_x = v_mc * cos(theta_mc) * cos(phi_mc)
                    v_y = v_mc * cos(theta_mc) * sin(phi_mc)
                    v_z = v_mc * sin(theta_mc)

                    # Get velocities in primed coordinate system
                    v_x_p = r_mat[0, 0] * v_x + r_mat[1, 0] * v_y + r_mat[2, 0] * v_z
                    v_y_p = r_mat[0, 1] * v_x + r_mat[1, 1] * v_y + r_mat[2, 1] * v_z
                    v_z_p = r_mat[0, 2] * v_x + r_mat[1, 2] * v_y + r_mat[2, 2] * v_z

                    # Elevation angle from the projection plane
                    alpha = asin(v_z_p / v_mc)

                    # Check if the velocity is within the v_lim and a_lim bounds
                    use_point = v_lim[0] <= v_z_p < v_lim[1]
                    use_point = use_point and (a_lim[0] <= alpha < a_lim[1])

                    # Check if the velocity along the x direction is outside of the grid
                    if v_x_p > vg_edges[-1] or v_x_p < vg_edges[0]:
                        continue

                    # Check if the velocity along the y direction is outside of the grid
                    if v_y_p > vg_edges[-1] or v_y_p < vg_edges[0]:
                        continue

                    # searchsorted returns the index where the element should be
                    # inserted to maintain order, so we subtract 1 to get the index of
                    # the bin that contains the element
                    i_vxg = np.searchsorted(vg_edges, v_x_p) - 1
                    i_vyg = np.searchsorted(vg_edges, v_y_p) - 1

                    # Special case for the first bin edge, which is the lower limit of
                    # the first bin and should be included in the first bin
                    if i_vxg == -1:
                        i_vxg = 0

                    if i_vyg == -1:
                        i_vyg = 0

                    if use_point:
                        f_g[i_vxg, i_vyg] += f_ijk * c_ijk / d_a_grid

    return f_g


@numba.jit(cache=True, nogil=True, parallel=True, nopython=True)
def _mc_cart_3d(
    vdf,
    v,
    phi,
    theta,
    d_v,
    d_v_m,
    d_phi,
    d_theta,
    vg_edges,
    d_a_grid,
    v_lim,
    a_lim,
    n_mc,
    r_mat,
):
    r"""Perform 3D Monte-Carlo interpolation of the VDFs

    Parameters
    ----------
    vdf : numpy.ndarray
        3D skymap particle velocity distribution function.
    v : numpy.ndarray
        1D array of instrument velocity bins centers.
    phi : numpy.ndarray
        1D array of instrument azimuthal angles bins centers.
    theta : numpy.ndarray
        1D array of instrument elevation angles bins centers.
    d_v : numpy.ndarray
        1D array of instrument velocity bins widths.
    d_v_m : numpy.ndarray
        1D array of minus velocity from bins centers.
    d_phi : numpy.ndarray
        1D array of instrument azimuthal angles bins widths.
    d_theta : numpy.ndarray
        1D array of instrument elevation angles bins widths.
    vg_egdes : double
        Bin centers of the velocity of the projection grid.
    d_a_grid : double
        Bin centers of the azimuthal angle of the projection in radians in
        the span [0,2*pi]. If this input is given, the projection will be 2D.
        If it is omitted, the projection will be 1D.
    v_lim : double
        Limits on the out-of-plane velocity interval in 2D and "transverse"
        velocity in 1D.
    a_lim : double
        Angular limit in degrees, can be combined with v_lim.
    n_mc : double
        Number of Monte-Carlo particle for the corresponding instrument bins.
    r_mat : double
        Frame transformation matrix.

    Returns
    -------
    f_g : double
        Reduced/interpolated distribution.

    """

    # Get dimension of the instrument and interpolation grid.
    n_v, n_ph, n_th = vdf.shape
    n_vg = len(vg_edges) - 1
    f_g = np.zeros((n_vg, n_vg, n_vg))

    for i in numba.prange(n_v):
        for j in range(n_ph):
            for k in range(n_th):
                n_mc_ijk = n_mc[i, j, k]

                if vdf[i][j][k] == 0.0:
                    continue

                dtau_ijk = v[i] ** 2 * cos(theta[k]) * d_v[i] * d_phi[j] * d_theta[k]
                c_ijk = dtau_ijk / n_mc_ijk
                f_ijk = vdf[i, j, k]

                for l_mc in range(n_mc_ijk):
                    # Generate Monte-Carlo particle speed, phi, and theta
                    if l_mc == 0:
                        # First Monte-Carlo particle is set at the bin center
                        d_v_mc = 0.0
                        d_phi_mc = 0.0
                    else:
                        d_v_mc = -random.random() * d_v[i] - d_v_m[0]
                        d_phi_mc = (random.random() - 0.5) * d_phi[j]

                    # convert instrument bin to cartesian velocity
                    v_mc = v[i] + d_v_mc
                    phi_mc = phi[j] + d_phi_mc

                    # This is needed to distribute the points evenly in space
                    # (more points further 'equatorward' since the slice is
                    # wider there).

                    if l_mc == 0:
                        theta_mc = theta[k]
                    else:
                        theta_1 = theta[k] - 0.5 * d_theta[k]
                        theta_2 = theta[k] + 0.5 * d_theta[k]
                        sin_theta_1 = sin(theta_1)
                        sin_theta_2 = sin(theta_2)
                        d_sin_theta = sin_theta_2 - sin_theta_1
                        sin_theta_mc = sin_theta_1 + random.random() * d_sin_theta
                        theta_mc = asin(sin_theta_mc)

                    v_x = v_mc * cos(theta_mc) * cos(phi_mc)
                    v_y = v_mc * cos(theta_mc) * sin(phi_mc)
                    v_z = v_mc * sin(theta_mc)

                    # Get velocities in primed coordinate system
                    # vxp = [vx, vy, vz] * xphat'; % all MC points
                    v_x_p = r_mat[0, 0] * v_x + r_mat[1, 0] * v_y + r_mat[2, 0] * v_z
                    v_y_p = r_mat[0, 1] * v_x + r_mat[1, 1] * v_y + r_mat[2, 1] * v_z
                    v_z_p = r_mat[0, 2] * v_x + r_mat[1, 2] * v_y + r_mat[2, 2] * v_z

                    alpha = asin(v_z_p / v_mc)

                    # Check if the velocity is within the v_lim and a_lim bounds
                    use_point = v_lim[0] <= v_z_p < v_lim[1]
                    use_point = use_point and (a_lim[0] <= alpha < a_lim[1])

                    if v_x_p > vg_edges[-1] or v_x_p < vg_edges[0]:
                        continue

                    if v_y_p > vg_edges[-1] or v_y_p < vg_edges[0]:
                        continue

                    if v_z_p > vg_edges[-1] or v_z_p < vg_edges[0]:
                        continue

                    # searchsorted returns the index where the element should be
                    # inserted to maintain order, so we subtract 1 to get the index of
                    # the bin that contains the element
                    i_vxg = np.searchsorted(vg_edges, v_x_p) - 1
                    i_vyg = np.searchsorted(vg_edges, v_y_p) - 1
                    i_vzg = np.searchsorted(vg_edges, v_z_p) - 1

                    # Special case for the last bin edge, which is the upper limit of
                    # the last bin and should be included in the last bin
                    if i_vxg == -1:
                        i_vxg = 0

                    if i_vyg == -1:
                        i_vyg = 0

                    if i_vzg == -1:
                        i_vzg = 0

                    if use_point:
                        f_g[i_vxg, i_vyg, i_vzg] += f_ijk * c_ijk / d_a_grid

    return f_g



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .calc_dt import calc_dt

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def gradient(inp):
    r"""Computes time derivative of the input variable.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time derivative of the input variable.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2017-07-18T13:03:34.000", "2017-07-18T13:07:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)

    Time derivative of the magnetic field

    >>> db_dt = pyrf.gradient(b_xyz)

    """

    # guess time step
    delta_t = calc_dt(inp)

    d_inp_dt = np.gradient(inp.data, axis=0) / delta_t

    out = xr.DataArray(
        d_inp_dt,
        coords=inp.coords,
        dims=inp.dims,
        attrs=inp.attrs,
    )

    if "UNITS" in out.attrs:
        out.attrs["UNITS"] += "/s"

    return out
